from unittest.mock import MagicMock, patch

import pytest

from scrapegraphai.graphs.abstract_graph import AbstractGraph
from scrapegraphai.graphs.depth_search_graph import DepthSearchGraph


class TestDepthSearchGraph:
    """Test suite for DepthSearchGraph class"""

    @pytest.mark.parametrize(
        "source, expected_input_key",
        [
            ("https://example.com", "url"),
            ("/path/to/local/directory", "local_dir"),
        ],
    )
    def test_depth_search_graph_initialization(self, source, expected_input_key):
        """
        Test that DepthSearchGraph initializes correctly with different source types.
        This test verifies that the input_key is set to 'url' for web sources and
        'local_dir' for local directory sources.
        """
        prompt = "Test prompt"
        config = {"llm": {"model": "mock_model"}}

        # Mock both BaseGraph and _create_llm method
        with (
            patch("scrapegraphai.graphs.depth_search_graph.BaseGraph"),
            patch.object(AbstractGraph, "_create_llm", return_value=MagicMock()),
        ):
            graph = DepthSearchGraph(prompt, source, config)

            assert graph.prompt == prompt
            assert graph.source == source
            assert graph.config == config
            assert graph.input_key == expected_input_key



"""
research_web module
"""

import re
from typing import List

import requests
from bs4 import BeautifulSoup
from langchain_community.tools import DuckDuckGoSearchResults


def search_on_web(
    query: str,
    search_engine: str = "duckduckgo",
    max_results: int = 10,
    port: int = 8080,
    timeout: int = 10,
    proxy: str | dict = None,
    serper_api_key: str = None,
    region: str = None,
    language: str = "en",
) -> List[str]:
    """Search web function with improved error handling and validation

    Args:
        query (str): Search query
        search_engine (str): Search engine to use
        max_results (int): Maximum number of results to return
        port (int): Port for SearXNG
        timeout (int): Request timeout in seconds
        proxy (str | dict): Proxy configuration
        serper_api_key (str): API key for Serper
        region (str): Country/region code (e.g., 'mx' for Mexico)
        language (str): Language code (e.g., 'es' for Spanish)
    """

    # Input validation
    if not query or not isinstance(query, str):
        raise ValueError("Query must be a non-empty string")

    search_engine = search_engine.lower()
    valid_engines = {"duckduckgo", "bing", "searxng", "serper"}
    if search_engine not in valid_engines:
        raise ValueError(f"Search engine must be one of: {', '.join(valid_engines)}")

    # Format proxy once
    formatted_proxy = None
    if proxy:
        formatted_proxy = format_proxy(proxy)

    try:
        results = []
        if search_engine == "duckduckgo":
            # Create a DuckDuckGo search object with max_results
            research = DuckDuckGoSearchResults(max_results=max_results)
            # Run the search
            res = research.run(query)
            # Extract URLs using regex
            results = re.findall(r"https?://[^\s,\]]+", res)

        elif search_engine == "bing":
            results = _search_bing(query, max_results, timeout, formatted_proxy)

        elif search_engine == "searxng":
            results = _search_searxng(query, max_results, port, timeout)

        elif search_engine == "serper":
            results = _search_serper(query, max_results, serper_api_key, timeout)

        return filter_pdf_links(results)

    except requests.Timeout:
        raise TimeoutError(f"Search request timed out after {timeout} seconds")
    except requests.RequestException as e:
        raise RuntimeError(f"Search request failed: {str(e)}")


def _search_bing(
    query: str, max_results: int, timeout: int, proxy: str = None
) -> List[str]:
    """Helper function for Bing search"""
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"
    }
    search_url = f"https://www.bing.com/search?q={query}"

    proxies = {"http": proxy, "https": proxy} if proxy else None
    response = requests.get(
        search_url, headers=headers, timeout=timeout, proxies=proxies
    )
    response.raise_for_status()

    soup = BeautifulSoup(response.text, "html.parser")
    return [
        result.find("a")["href"]
        for result in soup.find_all("li", class_="b_algo", limit=max_results)
    ]


def _search_searxng(query: str, max_results: int, port: int, timeout: int) -> List[str]:
    """Helper function for SearXNG search"""
    url = f"http://localhost:{port}/search"
    params = {
        "q": query,
        "format": "json",
        "engines": "google,duckduckgo,brave,qwant,bing",
    }
    response = requests.get(url, params=params, timeout=timeout)
    response.raise_for_status()
    return [
        result["url"] for result in response.json().get("results", [])[:max_results]
    ]


def _search_serper(
    query: str, max_results: int, serper_api_key: str, timeout: int
) -> List[str]:
    """Helper function for Serper API to get Google search results"""
    if not serper_api_key:
        raise ValueError("API key is required for Serper API")

    url = "https://google.serper.dev/search"
    payload = {"q": query, "num": max_results}

    headers = {"X-API-KEY": serper_api_key, "Content-Type": "application/json"}

    try:
        response = requests.post(
            url,
            headers=headers,
            json=payload,  # requests will handle JSON serialization
            timeout=timeout,
        )
        response.raise_for_status()

        # Extract only the organic search results
        results = response.json()
        organic_results = results.get("organic", [])
        urls = [result.get("link") for result in organic_results if result.get("link")]

        return urls[:max_results]

    except requests.exceptions.RequestException as e:
        raise RuntimeError(f"Serper API request failed: {str(e)}")


def format_proxy(proxy):
    if isinstance(proxy, dict):
        server = proxy.get("server")
        username = proxy.get("username")
        password = proxy.get("password")

        if all([username, password, server]):
            proxy_url = f"http://{username}:{password}@{server}"
            return proxy_url
        else:
            raise ValueError("Proxy dictionary is missing required fields.")
    elif isinstance(proxy, str):
        return proxy  # "https://username:password@ip:port"
    else:
        raise TypeError("Proxy should be a dictionary or a string.")


def filter_pdf_links(links: List[str]) -> List[str]:
    """
    Filters out any links that point to PDF files.

    Args:
        links (List[str]): A list of URLs as strings.

    Returns:
        List[str]: A list of URLs excluding any that end with '.pdf'.
    """
    return [link for link in links if not link.lower().endswith(".pdf")]



import asyncio
import sys
import time
from unittest.mock import ANY, AsyncMock, patch

import aiohttp
import pytest


class MockPlaywright:
    def __init__(self):
        self.chromium = AsyncMock()
        self.firefox = AsyncMock()


class MockBrowser:
    def __init__(self):
        self.new_context = AsyncMock()


class MockContext:
    def __init__(self):
        self.new_page = AsyncMock()


class MockPage:
    def __init__(self):
        self.goto = AsyncMock()
        self.wait_for_load_state = AsyncMock()
        self.content = AsyncMock()
        self.evaluate = AsyncMock()
        self.mouse = AsyncMock()
        self.mouse.wheel = AsyncMock()


@pytest.fixture
def mock_playwright():
    with patch("playwright.async_api.async_playwright") as mock:
        mock_pw = MockPlaywright()
        mock_browser = MockBrowser()
        mock_context = MockContext()
        mock_page = MockPage()

        mock_pw.chromium.launch.return_value = mock_browser
        mock_pw.firefox.launch.return_value = mock_browser
        mock_browser.new_context.return_value = mock_context
        mock_context.new_page.return_value = mock_page

        mock.return_value.__aenter__.return_value = mock_pw
        yield mock_pw, mock_browser, mock_context, mock_page


import pytest
from langchain_core.documents import Document

from scrapegraphai.docloaders.chromium import ChromiumLoader


async def dummy_scraper(url):
    """A dummy scraping function that returns dummy HTML content for the URL."""
    return f"<html>dummy content for {url}</html>"


@pytest.fixture
def loader_with_dummy(monkeypatch):
    """Fixture returning a ChromiumLoader instance with dummy scraping methods patched."""
    urls = ["http://example.com", "http://test.com"]
    loader = ChromiumLoader(urls, backend="playwright", requires_js_support=False)
    monkeypatch.setattr(loader, "ascrape_playwright", dummy_scraper)
    monkeypatch.setattr(loader, "ascrape_with_js_support", dummy_scraper)
    monkeypatch.setattr(loader, "ascrape_undetected_chromedriver", dummy_scraper)
    return loader


def test_lazy_load(loader_with_dummy):
    """Test that lazy_load yields Document objects with the correct dummy content and metadata."""
    docs = list(loader_with_dummy.lazy_load())
    assert len(docs) == 2
    for doc, url in zip(docs, loader_with_dummy.urls):
        assert isinstance(doc, Document)
        assert f"dummy content for {url}" in doc.page_content
        assert doc.metadata["source"] == url


@pytest.mark.asyncio
async def test_alazy_load(loader_with_dummy):
    """Test that alazy_load asynchronously yields Document objects with dummy content and proper metadata."""
    docs = [doc async for doc in loader_with_dummy.alazy_load()]
    assert len(docs) == 2
    for doc, url in zip(docs, loader_with_dummy.urls):
        assert isinstance(doc, Document)
        assert f"dummy content for {url}" in doc.page_content
        assert doc.metadata["source"] == url


@pytest.mark.asyncio
async def test_scrape_method_unsupported_backend():
    """Test that the scrape method raises a ValueError when an unsupported backend is provided."""
    loader = ChromiumLoader(["http://example.com"], backend="unsupported")
    with pytest.raises(ValueError):
        await loader.scrape("http://example.com")


@pytest.mark.asyncio
async def test_scrape_method_selenium(monkeypatch):
    """Test that the scrape method works correctly for selenium by returning the dummy selenium content."""

    async def dummy_selenium(url):
        return f"<html>dummy selenium content for {url}</html>"

    urls = ["http://example.com"]
    loader = ChromiumLoader(urls, backend="selenium")
    loader.browser_name = "chromium"
    monkeypatch.setattr(loader, "ascrape_undetected_chromedriver", dummy_selenium)
    result = await loader.scrape("http://example.com")
    assert "dummy selenium content" in result


@pytest.mark.asyncio
async def test_ascrape_playwright_scroll(mock_playwright):
    """Test the ascrape_playwright_scroll method with various configurations."""
    mock_pw, mock_browser, mock_context, mock_page = mock_playwright

    url = "http://example.com"
    loader = ChromiumLoader([url], backend="playwright")

    # Test with default parameters
    mock_page.evaluate.side_effect = [1000, 2000, 2000]  # Simulate scrolling
    result = await loader.ascrape_playwright_scroll(url)

    assert mock_page.goto.call_count == 1
    assert mock_page.wait_for_load_state.call_count == 1
    assert mock_page.mouse.wheel.call_count > 0
    assert mock_page.content.call_count == 1

    # Test with custom parameters
    mock_page.evaluate.side_effect = [1000, 2000, 3000, 4000, 4000]
    result = await loader.ascrape_playwright_scroll(
        url, timeout=10, scroll=10000, sleep=1, scroll_to_bottom=True
    )

    assert mock_page.goto.call_count == 2
    assert mock_page.wait_for_load_state.call_count == 2
    assert mock_page.mouse.wheel.call_count > 0
    assert mock_page.content.call_count == 2


@pytest.mark.asyncio
async def test_ascrape_with_js_support(mock_playwright):
    """Test the ascrape_with_js_support method with different browser configurations."""
    mock_pw, mock_browser, mock_context, mock_page = mock_playwright

    url = "http://example.com"
    loader = ChromiumLoader([url], backend="playwright", requires_js_support=True)

    # Test with Chromium
    result = await loader.ascrape_with_js_support(url, browser_name="chromium")
    assert mock_pw.chromium.launch.call_count == 1
    assert mock_page.goto.call_count == 1
    assert mock_page.content.call_count == 1

    # Test with Firefox
    result = await loader.ascrape_with_js_support(url, browser_name="firefox")
    assert mock_pw.firefox.launch.call_count == 1
    assert mock_page.goto.call_count == 2
    assert mock_page.content.call_count == 2

    # Test with invalid browser name
    with pytest.raises(ValueError):
        await loader.ascrape_with_js_support(url, browser_name="invalid")


@pytest.mark.asyncio
async def test_scrape_method_playwright(mock_playwright):
    """Test the scrape method with playwright backend."""
    mock_pw, mock_browser, mock_context, mock_page = mock_playwright

    url = "http://example.com"
    loader = ChromiumLoader([url], backend="playwright")

    mock_page.content.return_value = "<html>Playwright content</html>"
    result = await loader.scrape(url)

    assert "Playwright content" in result
    assert mock_pw.chromium.launch.call_count == 1
    assert mock_page.goto.call_count == 1
    assert mock_page.wait_for_load_state.call_count == 1
    assert mock_page.content.call_count == 1


@pytest.mark.asyncio
async def test_scrape_method_retry_logic(mock_playwright):
    """Test the retry logic in the scrape method."""
    mock_pw, mock_browser, mock_context, mock_page = mock_playwright

    url = "http://example.com"
    loader = ChromiumLoader([url], backend="playwright", retry_limit=3)

    # Simulate two failures and then a success
    mock_page.goto.side_effect = [asyncio.TimeoutError(), aiohttp.ClientError(), None]
    mock_page.content.return_value = "<html>Success after retries</html>"

    result = await loader.scrape(url)

    assert "Success after retries" in result
    assert mock_page.goto.call_count == 3
    assert mock_page.content.call_count == 1

    # Test failure after all retries
    mock_page.goto.side_effect = asyncio.TimeoutError()

    with pytest.raises(RuntimeError):
        await loader.scrape(url)

    assert mock_page.goto.call_count == 6  # 3 more attempts


@pytest.mark.asyncio
async def test_ascrape_playwright_scroll_invalid_params():
    """Test that ascrape_playwright_scroll raises ValueError for invalid scroll parameters."""
    loader = ChromiumLoader(["http://example.com"], backend="playwright")
    with pytest.raises(
        ValueError,
        match="If set, timeout value for scrolling scraper must be greater than 0.",
    ):
        await loader.ascrape_playwright_scroll("http://example.com", timeout=0)
    with pytest.raises(
        ValueError, match="Sleep for scrolling scraper value must be greater than 0."
    ):
        await loader.ascrape_playwright_scroll("http://example.com", sleep=0)
    with pytest.raises(
        ValueError,
        match="Scroll value for scrolling scraper must be greater than or equal to 5000.",
    ):
        await loader.ascrape_playwright_scroll("http://example.com", scroll=4000)


@pytest.mark.asyncio
async def test_ascrape_with_js_support_retry_failure(monkeypatch):
    """Test that ascrape_with_js_support retries and ultimately fails when page.goto always times out."""
    loader = ChromiumLoader(
        ["http://example.com"],
        backend="playwright",
        requires_js_support=True,
        retry_limit=2,
        timeout=1,
    )

    # Create dummy classes to simulate failure in page.goto
    class DummyPage:
        async def goto(self, url, wait_until):
            raise asyncio.TimeoutError("Forced timeout")

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>Dummy</html>"

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        async def new_context(self, **kwargs):
            return DummyContext()

        async def close(self):
            return

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    # Patch the async_playwright to return our dummy
    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())

    with pytest.raises(RuntimeError, match="Failed to scrape after"):
        await loader.ascrape_with_js_support("http://example.com")


@pytest.mark.asyncio
async def test_ascrape_undetected_chromedriver_success(monkeypatch):
    """Test that ascrape_undetected_chromedriver successfully returns content using the selenium backend."""
    # Create a dummy undetected_chromedriver module with a dummy Chrome driver.
    import types

    dummy_module = types.ModuleType("undetected_chromedriver")

    class DummyDriver:
        def __init__(self, options):
            self.options = options
            self.page_source = "<html>selenium content</html>"

        def quit(self):
            pass

    dummy_module.Chrome = lambda options: DummyDriver(options)
    monkeypatch.setitem(sys.modules, "undetected_chromedriver", dummy_module)

    urls = ["http://example.com"]
    loader = ChromiumLoader(urls, backend="selenium", retry_limit=1, timeout=5)
    loader.browser_name = "chromium"
    result = await loader.ascrape_undetected_chromedriver("http://example.com")
    assert "selenium content" in result


@pytest.mark.asyncio
async def test_lazy_load_exception(loader_with_dummy, monkeypatch):
    """Test that lazy_load propagates exception if the scraping function fails."""

    async def dummy_failure(url):
        raise Exception("Dummy scraping error")

    # Patch the scraping method to always raise an exception
    loader_with_dummy.backend = "playwright"
    monkeypatch.setattr(loader_with_dummy, "ascrape_playwright", dummy_failure)
    with pytest.raises(Exception, match="Dummy scraping error"):
        list(loader_with_dummy.lazy_load())


@pytest.mark.asyncio
async def test_ascrape_undetected_chromedriver_unsupported_browser(monkeypatch):
    """Test ascrape_undetected_chromedriver raises an error when an unsupported browser is provided."""
    import types

    dummy_module = types.ModuleType("undetected_chromedriver")
    # Provide a dummy Chrome; this will not be used for an unsupported browser.
    dummy_module.Chrome = lambda options: None
    monkeypatch.setitem(sys.modules, "undetected_chromedriver", dummy_module)

    loader = ChromiumLoader(
        ["http://example.com"], backend="selenium", retry_limit=1, timeout=1
    )
    loader.browser_name = "opera"  # Unsupported browser.
    with pytest.raises(UnboundLocalError):
        await loader.ascrape_undetected_chromedriver("http://example.com")


@pytest.mark.asyncio
async def test_alazy_load_partial_failure(monkeypatch):
    """Test that alazy_load propagates an exception if one of the scraping tasks fails."""
    urls = ["http://example.com", "http://fail.com"]
    loader = ChromiumLoader(urls, backend="playwright")

    async def partial_scraper(url):
        if "fail" in url:
            raise Exception("Scraping failed for " + url)
        return f"<html>Content for {url}</html>"

    monkeypatch.setattr(loader, "ascrape_playwright", partial_scraper)

    with pytest.raises(Exception, match="Scraping failed for http://fail.com"):
        [doc async for doc in loader.alazy_load()]


@pytest.mark.asyncio
async def test_ascrape_playwright_retry_failure(monkeypatch):
    """Test that ascrape_playwright retries scraping and raises RuntimeError after all attempts fail."""

    # Dummy classes to simulate persistent failure in page.goto for ascrape_playwright
    class DummyPage:
        async def goto(self, url, wait_until):
            raise asyncio.TimeoutError("Forced timeout in goto")

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>This should not be returned</html>"

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        async def new_context(self, **kwargs):
            return DummyContext()

        async def close(self):
            return

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())

    loader = ChromiumLoader(
        ["http://example.com"], backend="playwright", retry_limit=2, timeout=1
    )
    with pytest.raises(RuntimeError, match="Failed to scrape after 2 attempts"):
        await loader.ascrape_playwright("http://example.com")


@pytest.mark.asyncio
async def test_init_overrides():
    """Test that ChromiumLoader picks up and overrides attributes using kwargs."""
    urls = ["http://example.com"]
    loader = ChromiumLoader(
        urls,
        backend="playwright",
        headless=False,
        proxy={"http": "http://proxy"},
        load_state="load",
        requires_js_support=True,
        storage_state="state",
        browser_name="firefox",
        retry_limit=5,
        timeout=120,
        extra="value",
    )
    # Check that attributes are correctly set
    assert loader.headless is False
    assert loader.proxy == {"http": "http://proxy"}
    assert loader.load_state == "load"
    assert loader.requires_js_support is True
    assert loader.storage_state == "state"
    assert loader.browser_name == "firefox"
    assert loader.retry_limit == 5
    assert loader.timeout == 120
    # Check that extra kwargs go into browser_config
    assert loader.browser_config.get("extra") == "value"
    # Check that the backend remains as provided
    assert loader.backend == "playwright"


@pytest.mark.asyncio
async def test_lazy_load_with_js_support(monkeypatch):
    """Test that lazy_load uses ascrape_with_js_support when requires_js_support is True."""
    urls = ["http://example.com", "http://test.com"]
    loader = ChromiumLoader(urls, backend="playwright", requires_js_support=True)

    async def dummy_js(url):
        return f"<html>JS content for {url}</html>"

    monkeypatch.setattr(loader, "ascrape_with_js_support", dummy_js)
    docs = list(loader.lazy_load())
    assert len(docs) == 2
    for doc, url in zip(docs, urls):
        assert isinstance(doc, Document)
        assert f"JS content for {url}" in doc.page_content
        assert doc.metadata["source"] == url


@pytest.mark.asyncio
async def test_no_retry_returns_none(monkeypatch):
    """Test that ascrape_playwright returns None if retry_limit is set to 0."""
    urls = ["http://example.com"]
    loader = ChromiumLoader(urls, backend="playwright", retry_limit=0)

    # Even if we patch ascrape_playwright, the while loop won't run since retry_limit is 0, so it should return None.
    async def dummy(url, browser_name="chromium"):
        return f"<html>Content for {url}</html>"

    monkeypatch.setattr(loader, "ascrape_playwright", dummy)
    result = await loader.ascrape_playwright("http://example.com")
    # With retry_limit=0, the loop never runs and the function returns None.
    assert result is None


@pytest.mark.asyncio
async def test_alazy_load_empty_urls():
    """Test that alazy_load yields no documents when the urls list is empty."""
    loader = ChromiumLoader([], backend="playwright")
    docs = [doc async for doc in loader.alazy_load()]
    assert docs == []


def test_lazy_load_empty_urls():
    """Test that lazy_load yields no documents when the urls list is empty."""
    loader = ChromiumLoader([], backend="playwright")
    docs = list(loader.lazy_load())
    assert docs == []


@pytest.mark.asyncio
async def test_ascrape_undetected_chromedriver_missing_import(monkeypatch):
    """Test that ascrape_undetected_chromedriver raises ImportError when undetected_chromedriver is not installed."""
    # Remove undetected_chromedriver from sys.modules if it exists
    if "undetected_chromedriver" in sys.modules:
        monkeyatch_key = "undetected_chromedriver"
        monkeypatch.delenitem(sys.modules, monkeyatch_key)
    loader = ChromiumLoader(
        ["http://example.com"], backend="selenium", retry_limit=1, timeout=5
    )
    loader.browser_name = "chromium"
    with pytest.raises(
        ImportError, match="undetected_chromedriver is required for ChromiumLoader"
    ):
        await loader.ascrape_undetected_chromedriver("http://example.com")


@pytest.mark.asyncio
async def test_ascrape_undetected_chromedriver_quit_called(monkeypatch):
    """Test that ascrape_undetected_chromedriver calls driver.quit() on every attempt even when get() fails."""
    # List to collect each DummyDriver instance for later inspection.
    driver_instances = []
    attempt_counter = [0]

    class DummyDriver:
        def __init__(self, options):
            self.options = options
            self.quit_called = False
            driver_instances.append(self)

        def get(self, url):
            # Force a failure on the first attempt then succeed on subsequent attempts.
            if attempt_counter[0] < 1:
                attempt_counter[0] += 1
                raise aiohttp.ClientError("Forced failure")
            # If no failure, simply pass.

        @property
        def page_source(self):
            return "<html>driver content</html>"

        def quit(self):
            self.quit_called = True

    import types

    dummy_module = types.ModuleType("undetected_chromedriver")
    dummy_module.Chrome = lambda options: DummyDriver(options)
    monkeypatch.setitem(sys.modules, "undetected_chromedriver", dummy_module)

    urls = ["http://example.com"]
    loader = ChromiumLoader(urls, backend="selenium", retry_limit=2, timeout=5)
    loader.browser_name = "chromium"
    result = await loader.ascrape_undetected_chromedriver("http://example.com")
    assert "driver content" in result
    # Verify that two driver instances were used and that each had its quit() method called.
    assert len(driver_instances) == 2
    for driver in driver_instances:
        assert driver.quit_called is True


@pytest.mark.parametrize("backend", ["playwright", "selenium"])
def test_dynamic_import_failure(monkeypatch, backend):
    """Test that ChromiumLoader raises ImportError when dynamic_import fails."""

    def fake_dynamic_import(backend, message):
        raise ImportError("Test dynamic import error")

    monkeypatch.setattr(
        "scrapegraphai.docloaders.chromium.dynamic_import", fake_dynamic_import
    )
    with pytest.raises(ImportError, match="Test dynamic import error"):
        ChromiumLoader(["http://example.com"], backend=backend)


@pytest.mark.asyncio
async def test_ascrape_with_js_support_retry_success(monkeypatch):
    """Test that ascrape_with_js_support retries on failure and returns content on a subsequent successful attempt."""
    attempt_count = {"count": 0}

    class DummyPage:
        async def goto(self, url, wait_until):
            if attempt_count["count"] < 1:
                attempt_count["count"] += 1
                raise asyncio.TimeoutError("Forced timeout")
            # On second attempt, do nothing (simulate successful navigation)

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>Success on retry</html>"

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        async def new_context(self, **kwargs):
            return DummyContext()

        async def close(self):
            return

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())

    # Create a loader with JS support and a retry_limit of 2 (so one failure is allowed)
    loader = ChromiumLoader(
        ["http://example.com"],
        backend="playwright",
        requires_js_support=True,
        retry_limit=2,
        timeout=1,
    )
    result = await loader.ascrape_with_js_support("http://example.com")
    assert result == "<html>Success on retry</html>"


@pytest.mark.asyncio
async def test_proxy_parsing_in_init(monkeypatch):
    """Test that providing a proxy triggers the use of parse_or_search_proxy and sets loader.proxy correctly."""
    dummy_proxy_value = {"dummy": True}
    monkeypatch.setattr(
        "scrapegraphai.docloaders.chromium.parse_or_search_proxy",
        lambda proxy: dummy_proxy_value,
    )
    loader = ChromiumLoader(
        ["http://example.com"], backend="playwright", proxy="some_proxy_value"
    )
    assert loader.proxy == dummy_proxy_value


@pytest.mark.asyncio
async def test_scrape_method_selenium_firefox(monkeypatch):
    """Test that the scrape method works correctly for selenium with firefox backend."""

    async def dummy_selenium(url):
        return f"<html>dummy selenium firefox content for {url}</html>"

    urls = ["http://example.com"]
    loader = ChromiumLoader(urls, backend="selenium")
    loader.browser_name = "firefox"
    monkeypatch.setattr(loader, "ascrape_undetected_chromedriver", dummy_selenium)
    result = await loader.scrape("http://example.com")
    assert "dummy selenium firefox content" in result


def test_init_with_no_proxy():
    """Test that initializing ChromiumLoader with proxy=None results in loader.proxy being None."""
    urls = ["http://example.com"]
    loader = ChromiumLoader(urls, backend="playwright", proxy=None)
    assert loader.proxy is None


@pytest.mark.asyncio
async def test_ascrape_playwright_negative_retry(monkeypatch):
    """Test that ascrape_playwright returns None when retry_limit is negative (loop not executed)."""

    # Set-up a dummy playwright context which should never be used because retry_limit is negative.
    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                # Should not be called as retry_limit is negative.
                raise Exception("Should not launch browser")

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())
    urls = ["http://example.com"]
    loader = ChromiumLoader(urls, backend="playwright", retry_limit=-1)
    result = await loader.ascrape_playwright("http://example.com")
    assert result is None


@pytest.mark.asyncio
async def test_ascrape_with_js_support_negative_retry(monkeypatch):
    """Test that ascrape_with_js_support returns None when retry_limit is negative (loop not executed)."""

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                # Should not be called because retry_limit is negative.
                raise Exception("Should not launch browser")

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())
    urls = ["http://example.com"]
    loader = ChromiumLoader(
        urls, backend="playwright", requires_js_support=True, retry_limit=-1
    )
    try:
        result = await loader.ascrape_with_js_support("http://example.com")
    except RuntimeError:
        result = None
    assert result is None


@pytest.mark.asyncio
async def test_ascrape_with_js_support_storage_state(monkeypatch):
    """Test that ascrape_with_js_support passes the storage_state to the new_context call."""

    class DummyPage:
        async def goto(self, url, wait_until):
            return

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>Storage State Tested</html>"

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        def __init__(self):
            self.last_context_kwargs = None

        async def new_context(self, **kwargs):
            self.last_context_kwargs = kwargs
            return DummyContext()

        async def close(self):
            return

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                dummy_browser = DummyBrowser()
                dummy_browser.launch_kwargs = {
                    "headless": headless,
                    "proxy": proxy,
                    **kwargs,
                }
                return dummy_browser

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                dummy_browser = DummyBrowser()
                dummy_browser.launch_kwargs = {
                    "headless": headless,
                    "proxy": proxy,
                    **kwargs,
                }
                return dummy_browser

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())
    storage_state = "dummy_state"
    loader = ChromiumLoader(
        ["http://example.com"],
        backend="playwright",
        requires_js_support=True,
        storage_state=storage_state,
        retry_limit=1,
    )
    result = await loader.ascrape_with_js_support("http://example.com")
    # To ensure that new_context was called with the correct storage_state, we simulate a launch call
    browser = await DummyPW.chromium.launch(
        headless=loader.headless, proxy=loader.proxy
    )
    await browser.new_context(storage_state=loader.storage_state)
    assert browser.last_context_kwargs is not None
    assert browser.last_context_kwargs.get("storage_state") == storage_state
    assert "<html>Storage State Tested</html>" in result


@pytest.mark.asyncio
async def test_ascrape_playwright_browser_config(monkeypatch):
    """Test that ascrape_playwright passes extra browser_config kwargs to the browser launch."""
    captured_kwargs = {}

    class DummyPage:
        async def goto(self, url, wait_until):
            return

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>Config Tested</html>"

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        def __init__(self, config):
            self.config = config

        async def new_context(self, **kwargs):
            self.context_kwargs = kwargs
            return DummyContext()

        async def close(self):
            return

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                nonlocal captured_kwargs
                captured_kwargs = {"headless": headless, "proxy": proxy, **kwargs}
                return DummyBrowser(captured_kwargs)

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                nonlocal captured_kwargs
                captured_kwargs = {"headless": headless, "proxy": proxy, **kwargs}
                return DummyBrowser(captured_kwargs)

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())
    extra_kwarg_value = "test_value"
    loader = ChromiumLoader(
        ["http://example.com"],
        backend="playwright",
        extra=extra_kwarg_value,
        retry_limit=1,
    )
    result = await loader.ascrape_playwright("http://example.com")
    assert captured_kwargs.get("extra") == extra_kwarg_value
    assert "<html>Config Tested</html>" in result


@pytest.mark.asyncio
async def test_scrape_method_js_support(monkeypatch):
    """Test that scrape method calls ascrape_with_js_support when requires_js_support is True."""

    async def dummy_js(url):
        return f"<html>JS supported content for {url}</html>"

    urls = ["http://example.com"]
    loader = ChromiumLoader(urls, backend="playwright", requires_js_support=True)
    monkeypatch.setattr(loader, "ascrape_with_js_support", dummy_js)
    result = await loader.scrape("http://example.com")
    assert "JS supported content" in result


@pytest.mark.asyncio
async def test_ascrape_playwright_scroll_retry_failure(monkeypatch):
    """Test that ascrape_playwright_scroll retries on failure and returns an error message after retry_limit attempts."""

    # Dummy page that always raises Timeout on goto
    class DummyPage:
        async def goto(self, url, wait_until):
            raise asyncio.TimeoutError("Simulated timeout in goto")

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>No Content</html>"

        evaluate = AsyncMock(
            side_effect=asyncio.TimeoutError("Simulated timeout in evaluate")
        )

        mouse = AsyncMock()

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        async def new_context(self, **kwargs):
            return DummyContext()

        async def close(self):
            return

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())

    urls = ["http://example.com"]
    loader = ChromiumLoader(urls, backend="playwright", retry_limit=2, timeout=1)
    # Use a scroll value just above minimum and a sleep value > 0
    result = await loader.ascrape_playwright_scroll(
        "http://example.com", scroll=5000, sleep=1
    )
    assert "Error: Network error after 2 attempts" in result


@pytest.mark.asyncio
async def test_alazy_load_order(monkeypatch):
    """Test that alazy_load returns documents in the same order as the input URLs even if scraping tasks complete out of order."""
    urls = [
        "http://example.com/first",
        "http://example.com/second",
        "http://example.com/third",
    ]
    loader = ChromiumLoader(urls, backend="playwright")

    async def delayed_scraper(url):
        # Delay inversely proportional to a function of the url to scramble finish order
        import asyncio

        delay = 0.3 - 0.1 * (len(url) % 3)
        await asyncio.sleep(delay)
        return f"<html>Content for {url}</html>"

    monkeypatch.setattr(loader, "ascrape_playwright", delayed_scraper)

    docs = [doc async for doc in loader.alazy_load()]
    # Ensure that the order of documents matches the order of input URLs
    for doc, url in zip(docs, urls):
        assert doc.metadata["source"] == url
        assert f"Content for {url}" in doc.page_content


@pytest.mark.asyncio
async def test_ascrape_with_js_support_calls_close(monkeypatch):
    """Test that ascrape_with_js_support calls browser.close() after scraping."""
    close_called_flag = {"called": False}

    class DummyPage:
        async def goto(self, url, wait_until):
            return

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>Dummy Content</html>"

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        async def new_context(self, **kwargs):
            return DummyContext()

        async def close(self):
            close_called_flag["called"] = True
            return

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())

    urls = ["http://example.com"]
    loader = ChromiumLoader(
        urls, backend="playwright", requires_js_support=True, retry_limit=1, timeout=5
    )
    result = await loader.ascrape_with_js_support("http://example.com")
    assert result == "<html>Dummy Content</html>"
    assert close_called_flag["called"] is True


@pytest.mark.asyncio
async def test_lazy_load_invalid_backend(monkeypatch):
    """Test that lazy_load raises AttributeError if the scraping method for an invalid backend is missing."""
    # Create a loader instance with a backend that does not have a corresponding scraping method.
    loader = ChromiumLoader(["http://example.com"], backend="nonexistent")
    with pytest.raises(AttributeError):
        # lazy_load calls asyncio.run(scraping_fn(url)) for each URL.
        list(loader.lazy_load())


@pytest.mark.asyncio
async def test_ascrape_undetected_chromedriver_failure(monkeypatch):
    """Test that ascrape_undetected_chromedriver returns an error message after all retry attempts when driver.get always fails."""
    import types

    # Create a dummy undetected_chromedriver module with a dummy Chrome driver that always fails.
    dummy_module = types.ModuleType("undetected_chromedriver")

    class DummyDriver:
        def __init__(self, options):
            self.options = options
            self.quit_called = False

        def get(self, url):
            # Simulate a failure in fetching the page.
            raise aiohttp.ClientError("Forced failure in get")

        @property
        def page_source(self):
            return "<html>This should not be reached</html>"

        def quit(self):
            self.quit_called = True

    dummy_module.Chrome = lambda options: DummyDriver(options)
    monkeypatch.setitem(sys.modules, "undetected_chromedriver", dummy_module)

    loader = ChromiumLoader(
        ["http://example.com"], backend="selenium", retry_limit=2, timeout=1
    )
    loader.browser_name = "chromium"
    result = await loader.ascrape_undetected_chromedriver("http://example.com")
    # Check that the error message indicates the number of attempts and the forced failure.
    assert "Error: Network error after 2 attempts" in result


@pytest.mark.asyncio
async def test_ascrape_playwright_scroll_constant_height(mock_playwright):
    """Test that ascrape_playwright_scroll exits the scroll loop when page height remains constant."""
    mock_pw, mock_browser, mock_context, mock_page = mock_playwright
    # Set evaluate to always return constant height value (simulate constant page height)
    mock_page.evaluate.return_value = 1000
    # Return dummy content once scrolling loop breaks
    mock_page.content.return_value = "<html>Constant height content</html>"
    # Use a scroll value above minimum and a very short sleep to cycle quickly
    loader = ChromiumLoader(["http://example.com"], backend="playwright")
    result = await loader.ascrape_playwright_scroll(
        "http://example.com", scroll=6000, sleep=0.1
    )
    assert "Constant height content" in result


def test_lazy_load_empty_content(monkeypatch):
    """Test that lazy_load yields a Document with empty content if the scraper returns an empty string."""
    from langchain_core.documents import Document

    urls = ["http://example.com"]
    loader = ChromiumLoader(urls, backend="playwright", requires_js_support=False)

    async def dummy_scraper(url):
        return ""

    monkeypatch.setattr(loader, "ascrape_playwright", dummy_scraper)
    docs = list(loader.lazy_load())
    assert len(docs) == 1
    for doc in docs:
        assert isinstance(doc, Document)
        assert doc.page_content == ""
        assert doc.metadata["source"] in urls


@pytest.mark.asyncio
async def test_lazy_load_scraper_returns_none(monkeypatch):
    """Test that lazy_load yields Document objects with page_content as None when the scraper returns None."""
    urls = ["http://example.com", "http://test.com"]
    loader = ChromiumLoader(urls, backend="playwright")

    async def dummy_none(url):
        return None

    monkeypatch.setattr(loader, "ascrape_playwright", dummy_none)
    docs = list(loader.lazy_load())
    assert len(docs) == 2
    for doc, url in zip(docs, urls):
        from langchain_core.documents import Document

        assert isinstance(doc, Document)
        assert doc.page_content is None
        assert doc.metadata["source"] == url


@pytest.mark.asyncio
async def test_alazy_load_mixed_none_and_content(monkeypatch):
    """Test that alazy_load yields Document objects in order when one scraper returns None and the other valid HTML."""
    urls = ["http://example.com", "http://none.com"]
    loader = ChromiumLoader(urls, backend="playwright")

    async def mixed_scraper(url):
        if "none" in url:
            return None
        return f"<html>Valid content for {url}</html>"

    monkeypatch.setattr(loader, "ascrape_playwright", mixed_scraper)
    docs = [doc async for doc in loader.alazy_load()]
    assert len(docs) == 2
    # Ensure order is preserved and check contents
    assert docs[0].metadata["source"] == "http://example.com"
    assert "<html>Valid content for http://example.com</html>" in docs[0].page_content
    assert docs[1].metadata["source"] == "http://none.com"
    assert docs[1].page_content is None


@pytest.mark.asyncio
async def test_ascrape_with_js_support_exception_cleanup(monkeypatch):
    """Test that ascrape_with_js_support calls browser.close() after an exception occurs."""
    close_called_flag = {"called": False}

    class DummyPage:
        async def goto(self, url, wait_until):
            raise asyncio.TimeoutError("Forced timeout")

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>No Content</html>"

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        async def new_context(self, **kwargs):
            return DummyContext()

        async def close(self):
            close_called_flag["called"] = True
            return

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())

    loader = ChromiumLoader(
        ["http://example.com"],
        backend="playwright",
        requires_js_support=True,
        retry_limit=1,
        timeout=1,
    )

    with pytest.raises(RuntimeError, match="Failed to scrape after 1 attempts"):
        await loader.ascrape_with_js_support("http://example.com")


@patch("scrapegraphai.docloaders.chromium.dynamic_import")
def test_init_dynamic_import_called(mock_dynamic_import):
    """Test that dynamic_import is called during initialization."""
    urls = ["http://example.com"]
    _ = ChromiumLoader(urls, backend="playwright")
    mock_dynamic_import.assert_called_with("playwright", ANY)


@pytest.mark.asyncio
async def test_alazy_load_selenium_backend(monkeypatch):
    """Test that alazy_load correctly yields Document objects when using selenium backend."""
    urls = ["http://example.com", "http://selenium.com"]
    loader = ChromiumLoader(urls, backend="selenium")

    async def dummy_selenium(url):
        return f"<html>dummy selenium backend content for {url}</html>"

    monkeypatch.setattr(loader, "ascrape_undetected_chromedriver", dummy_selenium)
    docs = [doc async for doc in loader.alazy_load()]
    for doc, url in zip(docs, urls):
        assert f"dummy selenium backend content for {url}" in doc.page_content
        assert doc.metadata["source"] == url
    assert close_called_flag["called"] is True


@pytest.mark.asyncio
async def test_ascrape_undetected_chromedriver_zero_retry(monkeypatch):
    """Test that ascrape_undetected_chromedriver returns empty result when retry_limit is set to 0."""
    import types

    # Create a dummy undetected_chromedriver module where Chrome is defined but will not be used.
    dummy_module = types.ModuleType("undetected_chromedriver")
    dummy_module.Chrome = lambda options: None
    monkeypatch.setitem(sys.modules, "undetected_chromedriver", dummy_module)

    loader = ChromiumLoader(
        ["http://example.com"], backend="selenium", retry_limit=0, timeout=5
    )
    loader.browser_name = "chromium"
    # With retry_limit=0, the while loop never runs so the result remains an empty string.
    result = await loader.ascrape_undetected_chromedriver("http://example.com")
    assert result == ""


@pytest.mark.asyncio
async def test_scrape_selenium_exception(monkeypatch):
    """Test that the scrape method for selenium backend raises a ValueError when ascrape_undetected_chromedriver fails."""

    async def failing_scraper(url):
        raise Exception("dummy error")

    urls = ["http://example.com"]
    loader = ChromiumLoader(urls, backend="selenium", retry_limit=1, timeout=5)
    loader.browser_name = "chromium"
    monkeypatch.setattr(loader, "ascrape_undetected_chromedriver", failing_scraper)
    with pytest.raises(
        ValueError, match="Failed to scrape with undetected chromedriver: dummy error"
    ):
        await loader.scrape("http://example.com")


@pytest.mark.asyncio
async def test_ascrape_playwright_scroll_exception_cleanup(monkeypatch):
    """Test that ascrape_playwright_scroll calls browser.close() when an exception occurs during page navigation."""
    close_called = {"called": False}

    class DummyPage:
        async def goto(self, url, wait_until):
            raise asyncio.TimeoutError("Simulated timeout in goto")

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>Never reached</html>"

        async def evaluate(self, script):
            return 1000  # constant height value to simulate no progress in scrolling

        mouse = AsyncMock()
        mouse.wheel = AsyncMock()

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        async def new_context(self, **kwargs):
            return DummyContext()

        async def close(self):
            close_called["called"] = True

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())

    loader = ChromiumLoader(
        ["http://example.com"],
        backend="playwright",
        retry_limit=2,
        timeout=1,
        headless=True,
    )
    result = await loader.ascrape_playwright_scroll(
        "http://example.com", scroll=5000, sleep=0.1, scroll_to_bottom=True
    )

    assert "Error: Network error after" in result
    assert close_called["called"] is True


@pytest.mark.asyncio
async def test_ascrape_with_js_support_non_timeout_retry(monkeypatch):
    """Test that ascrape_with_js_support retries on a non-timeout exception and eventually succeeds."""
    attempt = {"count": 0}

    class DummyPage:
        async def goto(self, url, wait_until):
            if attempt["count"] < 1:
                attempt["count"] += 1
                raise ValueError("Non-timeout error")

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>Success after non-timeout retry</html>"

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        async def new_context(self, **kwargs):
            return DummyContext()

        async def close(self):
            return

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())
    loader = ChromiumLoader(
        ["http://nontimeout.com"],
        backend="playwright",
        requires_js_support=True,
        retry_limit=2,
        timeout=1,
    )
    result = await loader.ascrape_with_js_support("http://nontimeout.com")
    assert "Success after non-timeout retry" in result


@pytest.mark.asyncio
async def test_scrape_uses_js_support_flag(monkeypatch):
    """Test that the scrape method uses ascrape_with_js_support when requires_js_support is True."""

    async def dummy_js(url, browser_name="chromium"):
        return f"<html>JS flag content for {url}</html>"

    async def dummy_playwright(url, browser_name="chromium"):
        return f"<html>Playwright content for {url}</html>"

    urls = ["http://example.com"]
    loader = ChromiumLoader(urls, backend="playwright", requires_js_support=True)
    monkeypatch.setattr(loader, "ascrape_with_js_support", dummy_js)
    monkeypatch.setattr(loader, "ascrape_playwright", dummy_playwright)
    result = await loader.scrape("http://example.com")
    assert "JS flag content" in result


@pytest.mark.asyncio
async def test_ascrape_playwright_calls_apply_stealth(monkeypatch):
    """Test that ascrape_playwright calls Malenia.apply_stealth on the browser context."""
    flag = {"applied": False}

    async def dummy_apply_stealth(context):
        flag["applied"] = True

    monkeypatch.setattr(
        "scrapegraphai.docloaders.chromium.Malenia.apply_stealth", dummy_apply_stealth
    )

    class DummyPage:
        async def goto(self, url, wait_until):
            return

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>Stealth Applied Content</html>"

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        async def new_context(self, **kwargs):
            return DummyContext()

        async def close(self):
            return

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())

    loader = ChromiumLoader(["http://example.com"], backend="playwright")
    result = await loader.ascrape_playwright("http://example.com")
    assert flag["applied"] is True
    assert "Stealth Applied Content" in result


@pytest.mark.asyncio
async def test_lazy_load_non_string_scraper(monkeypatch):
    """Test that lazy_load yields Document objects even if the scraping function returns a non‐string value."""
    urls = ["http://example.com"]
    loader = ChromiumLoader(urls, backend="playwright", requires_js_support=False)

    async def dummy_non_string(url):
        # Return an integer instead of an HTML string
        return 12345

    monkeypatch.setattr(loader, "ascrape_playwright", dummy_non_string)
    docs = list(loader.lazy_load())
    # Check that we get one Document and its page_content is the non‐string value returned by the scraper
    from langchain_core.documents import Document

    assert len(docs) == 1
    for doc in docs:
        assert isinstance(doc, Document)
        assert doc.page_content == 12345
        assert doc.metadata["source"] in urls


@pytest.mark.asyncio
async def test_alazy_load_non_string_scraper(monkeypatch):
    """Test that alazy_load yields Document objects with a non‐string page_content when the JS scraping function returns a non‐string value."""
    urls = ["http://nonstring.com"]
    # Instantiate loader with requires_js_support True so that alazy_load calls ascrape_with_js_support
    loader = ChromiumLoader(urls, backend="playwright", requires_js_support=True)

    # Define a dummy scraper that returns an integer (non‐string)
    async def dummy_non_string(url, browser_name="chromium"):
        return 54321

    monkeypatch.setattr(loader, "ascrape_with_js_support", dummy_non_string)
    docs = [doc async for doc in loader.alazy_load()]
    from langchain_core.documents import Document

    assert len(docs) == 1
    assert isinstance(docs[0], Document)
    assert docs[0].page_content == 54321
    assert docs[0].metadata["source"] == "http://nonstring.com"


@pytest.mark.asyncio
async def test_ascrape_playwright_scroll_timeout_none(monkeypatch, mock_playwright):
    """Test ascrape_playwright_scroll when timeout is None and scroll_to_bottom is True.
    The test uses a dummy page.evaluate sequence to simulate increasing then constant page height.
    """
    mock_pw, mock_browser, mock_context, mock_page = mock_playwright
    # Simulate a first scroll returns 1000, then 2000, then constant height (2000)
    mock_page.evaluate.side_effect = [1000, 2000, 2000, 2000, 2000]
    # When scrolling is done the final content is returned
    mock_page.content.return_value = "<html>Timeout None Content</html>"
    loader = ChromiumLoader(["http://example.com"], backend="playwright")
    result = await loader.ascrape_playwright_scroll(
        "http://example.com",
        timeout=None,
        scroll=6000,
        sleep=0.1,
        scroll_to_bottom=True,
    )
    assert "timeout none content" in result.lower()


@pytest.mark.asyncio
async def test_ascrape_with_js_support_browser_error_cleanup(monkeypatch):
    """Test ascrape_with_js_support to ensure that browser.close() is always called even if an exception occurs.
    This simulates a navigation error and checks that on exception the browser is properly closed.
    """
    close_called = {"called": False}

    class DummyPage:
        async def goto(self, url, wait_until):
            raise aiohttp.ClientError("Navigation error")

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>Error Content</html>"

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        async def new_context(self, **kwargs):
            return DummyContext()

        async def close(self):
            close_called["called"] = True

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())
    loader = ChromiumLoader(
        ["http://example.com"],
        backend="playwright",
        requires_js_support=True,
        retry_limit=1,
        timeout=1,
    )
    with pytest.raises(RuntimeError):
        await loader.ascrape_with_js_support("http://example.com")
    assert close_called["called"] is True


def dummy_non_async_scraper(url):
    """A dummy scraper function that is not asynchronous."""
    return "non-async result"


def test_lazy_load_with_non_async_scraper(monkeypatch, loader_with_dummy):
    """Test that lazy_load raises a ValueError when a non-async function is used as the scraper.
    In this case, using a non-async function in place of an async scraper should lead to a ValueError.
    """
    monkeypatch.setattr(
        loader_with_dummy, "ascrape_playwright", dummy_non_async_scraper
    )
    with pytest.raises(
        ValueError, match="a coroutine was expected, got 'non-async result'"
    ):
        list(loader_with_dummy.lazy_load())


@pytest.mark.asyncio
async def test_ascrape_playwright_stealth_exception_cleanup(monkeypatch):
    """Test that ascrape_playwright calls browser.close() even if Malenia.apply_stealth fails."""
    fail_flag = {"closed": False}

    class DummyPage:
        async def goto(self, url, wait_until):
            return

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>Content</html>"

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        async def new_context(self, **kwargs):
            return DummyContext()

        async def close(self):
            fail_flag["closed"] = True

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())

    async def fail_apply_stealth(context):
        raise ValueError("Stealth failed")

    monkeypatch.setattr(
        "scrapegraphai.docloaders.chromium.Malenia.apply_stealth", fail_apply_stealth
    )

    loader = ChromiumLoader(
        ["http://example.com"], backend="playwright", retry_limit=1, timeout=1
    )
    with pytest.raises(RuntimeError, match="Failed to scrape after 1 attempts"):
        await loader.ascrape_playwright("http://example.com")
    assert fail_flag["closed"] is True


@pytest.mark.asyncio
async def test_ascrape_with_js_support_value_error_success(monkeypatch):
    """Test that ascrape_with_js_support retries on ValueError and eventually succeeds."""
    attempt_count = {"count": 0}

    class DummyPage:
        async def goto(self, url, wait_until):
            if attempt_count["count"] < 1:
                attempt_count["count"] += 1
                raise ValueError("Test value error")
            return

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>Success after ValueError</html>"

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        async def new_context(self, **kwargs):
            return DummyContext()

        async def close(self):
            return

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())
    loader = ChromiumLoader(
        ["http://example.com"],
        backend="playwright",
        requires_js_support=True,
        retry_limit=2,
        timeout=1,
    )
    result = await loader.ascrape_with_js_support("http://example.com")
    assert "Success after ValueError" in result


@pytest.mark.asyncio
async def test_ascrape_with_js_support_value_error_failure(monkeypatch):
    """Test that ascrape_with_js_support raises RuntimeError after exhausting retries on persistent ValueError."""

    class DummyPage:
        async def goto(self, url, wait_until):
            raise ValueError("Persistent value error")

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>Should not reach here</html>"

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        async def new_context(self, **kwargs):
            return DummyContext()

        async def close(self):
            return

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())
    loader = ChromiumLoader(
        ["http://example.com"],
        backend="playwright",
        requires_js_support=True,
        retry_limit=1,
        timeout=1,
    )
    with pytest.raises(RuntimeError, match="Failed to scrape after 1 attempts"):
        await loader.ascrape_with_js_support("http://example.com")


@pytest.mark.asyncio
async def test_ascrape_playwright_scroll_scroll_to_bottom_false(
    monkeypatch, mock_playwright
):
    """Test ascrape_playwright_scroll with scroll_to_bottom=False.
    Simulate a page whose scroll height increases initially then remains constant;
    with a short timeout the function should break and return the page content.
    """
    mock_pw, mock_browser, mock_context, mock_page = mock_playwright
    # simulate a sequence of scroll heights: first increases then remains constant
    mock_page.evaluate.side_effect = [1000, 1500, 1500, 1500, 1500]
    mock_page.content.return_value = (
        "<html>Timeout reached without scrolling bottom</html>"
    )

    # Create loader with default load_state and short timeout such that the loop terminates
    loader = ChromiumLoader(
        ["http://example.com"], backend="playwright", load_state="domcontentloaded"
    )
    result = await loader.ascrape_playwright_scroll(
        "http://example.com", timeout=1, scroll=6000, sleep=0.1, scroll_to_bottom=False
    )
    assert "Timeout reached" in result


@pytest.mark.asyncio
async def test_ascrape_with_js_support_browser_name_override_new(monkeypatch):
    """Test that ascrape_with_js_support calls the firefox branch correctly when browser_name is set to "firefox".
    This simulates a dummy playwright that returns a DummyBrowser and content when using firefox.
    """

    class DummyPage:
        async def goto(self, url, wait_until):
            return

        async def wait_for_load_state(self, state):
            return

        async def content(self):
            return "<html>Firefox content</html>"

    class DummyContext:
        async def new_page(self):
            return DummyPage()

    class DummyBrowser:
        async def new_context(self, **kwargs):
            self.context_kwargs = kwargs
            return DummyContext()

        async def close(self):
            return

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                raise Exception("Chromium branch not used for this test")

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())
    loader = ChromiumLoader(
        ["http://example.com"], backend="playwright", requires_js_support=True
    )
    result = await loader.ascrape_with_js_support(
        "http://example.com", browser_name="firefox"
    )
    assert "Firefox content" in result


@pytest.mark.asyncio
async def test_ascrape_playwright_scroll_load_state(mock_playwright):
    """Test that ascrape_playwright_scroll waits for the custom load_state value."""
    mock_pw, mock_browser, mock_context, mock_page = mock_playwright
    url = "http://example.com"
    # Instantiate the loader with a non-default load_state ("custom_state")
    loader = ChromiumLoader([url], backend="playwright", load_state="custom_state")

    # Simulate constant page height so that scrolling stops.
    # First call returns 1000 then remains constant.
    mock_page.evaluate.side_effect = [1000, 1000]
    mock_page.content.return_value = "<html>Done</html>"

    result = await loader.ascrape_playwright_scroll(
        url, timeout=1, scroll=5000, sleep=0.1, scroll_to_bottom=True
    )

    # Check that wait_for_load_state was called with the custom load_state value.
    mock_page.wait_for_load_state.assert_called_with("custom_state")
    assert "Done" in result


@pytest.mark.asyncio
async def test_alazy_load_concurrency(monkeypatch):
    """Test that alazy_load runs tasks concurrently by measuring elapsed time.
    Each dummy task sleeps for 0.5 seconds. If run sequentially the total time
    would be at least 1.5 seconds for three URLs. Running concurrently should be
    significantly faster.
    """
    import time

    urls = ["http://example.com/1", "http://example.com/2", "http://example.com/3"]
    loader = ChromiumLoader(urls, backend="playwright")

    async def dummy_delay(url):
        await asyncio.sleep(0.5)
        return f"<html>Content for {url}</html>"

    monkeypatch.setattr(loader, "ascrape_playwright", dummy_delay)
    start_time = time.monotonic()
    docs = [doc async for doc in loader.alazy_load()]
    elapsed = time.monotonic() - start_time
    # In sequential execution elapsed time would be at least 1.5s;
    # if tasks run concurrently it should be considerably less.
    assert elapsed < 1.0, f"Expected concurrent execution but took {elapsed} seconds"
    for doc, url in zip(docs, urls):
        assert url in doc.metadata["source"]
        assert f"Content for {url}" in doc.page_content


@pytest.mark.asyncio
async def test_scrape_playwright_value_error_retry_failure(monkeypatch):
    """Test that ascrape_playwright retries on ValueError and ultimately raises RuntimeError after exhausting retries."""

    async def always_value_error(url, browser_name="chromium"):
        raise ValueError("Forced value error")

    urls = ["http://example.com"]
    # requires_js_support is False so that scraper calls ascrape_playwright.
    loader = ChromiumLoader(
        urls, backend="playwright", requires_js_support=False, retry_limit=2, timeout=1
    )
    monkeypatch.setattr(loader, "ascrape_playwright", always_value_error)
    with pytest.raises(RuntimeError, match="Failed to scrape after 2 attempts"):
        await loader.scrape("http://example.com")


@pytest.mark.asyncio
async def test_invalid_proxy_raises_error(monkeypatch):
    """Test that providing an invalid proxy causes a ValueError during initialization (via parse_or_search_proxy)."""

    def fake_parse_or_search_proxy(proxy):
        raise ValueError("Invalid proxy")

    monkeypatch.setattr(
        "scrapegraphai.docloaders.chromium.parse_or_search_proxy",
        fake_parse_or_search_proxy,
    )
    with pytest.raises(ValueError, match="Invalid proxy"):
        ChromiumLoader(["http://example.com"], backend="playwright", proxy="bad_proxy")


@pytest.mark.asyncio
async def test_alazy_load_with_single_url_string(monkeypatch):
    """Test that alazy_load yields Document objects when urls is a string (iterating over characters)."""
    # Passing a string as URL; lazy_load will iterate each character.
    loader = ChromiumLoader(
        "http://example.com", backend="playwright", requires_js_support=False
    )

    async def dummy_scraper(url, browser_name="chromium"):
        return f"<html>{url}</html>"

    monkeypatch.setattr(loader, "ascrape_playwright", dummy_scraper)
    docs = [doc async for doc in loader.alazy_load()]
    # The expected number of documents is the length of the string
    expected_length = len("http://example.com")
    assert len(docs) == expected_length
    # Check that the first document’s source is the first character ('h')
    assert docs[0].metadata["source"] == "h"


def test_lazy_load_with_single_url_string(monkeypatch):
    """Test that lazy_load yields Document objects when urls is a string (iterating over characters)."""
    loader = ChromiumLoader(
        "http://example.com", backend="playwright", requires_js_support=False
    )

    async def dummy_scraper(url, browser_name="chromium"):
        return f"<html>{url}</html>"

    monkeypatch.setattr(loader, "ascrape_playwright", dummy_scraper)
    docs = list(loader.lazy_load())
    expected_length = len("http://example.com")
    assert len(docs) == expected_length
    # The first character from the URL is 'h'
    assert docs[0].metadata["source"] == "h"


@pytest.mark.asyncio
async def test_ascrape_playwright_scroll_invalid_type(monkeypatch):
    """Test that ascrape_playwright_scroll raises TypeError when invalid types are passed for scroll or sleep."""
    # Create a dummy playwright so that evaluate and content can be called

    loader = ChromiumLoader(["http://example.com"], backend="playwright")
    # Passing a non‐numeric sleep value should eventually trigger an error
    with pytest.raises(TypeError):
        await loader.ascrape_playwright_scroll(
            "http://example.com", scroll=6000, sleep="2", scroll_to_bottom=False
        )


@pytest.mark.asyncio
async def test_alazy_load_non_iterable_urls():
    """Test that alazy_load raises TypeError when urls is not an iterable (e.g., integer)."""
    with pytest.raises(TypeError):
        # Passing an integer as urls should cause a TypeError during iteration.
        loader = ChromiumLoader(123, backend="playwright")
        [doc async for doc in loader.alazy_load()]


def test_lazy_load_non_iterable_urls():
    """Test that lazy_load raises TypeError when urls is not an iterable (e.g., integer)."""
    with pytest.raises(TypeError):
        loader = ChromiumLoader(456, backend="playwright")

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())

    # Create a loader instance with retry_limit=2 so that one failure is allowed.


@pytest.mark.asyncio
async def test_ascrape_playwright_caplog(monkeypatch, caplog):
    """
    Test that ascrape_playwright recovers on failure and that error messages are logged.
    This test simulates one failed attempt (via a Timeout) and then a successful attempt.
    """
    # Create a loader instance with a retry limit of 2 and a short timeout.
    loader = ChromiumLoader(
        ["http://example.com"], backend="playwright", retry_limit=2, timeout=1
    )
    attempt = {"count": 0}

    async def dummy_ascrape(url, browser_name="chromium"):
        if attempt["count"] < 1:
            attempt["count"] += 1
            raise asyncio.TimeoutError("Simulated Timeout")
        return "Recovered Content"

    monkeypatch.setattr(loader, "ascrape_playwright", dummy_ascrape)
    with caplog.at_level("ERROR"):
        result = await loader.ascrape_playwright("http://example.com")
    assert "Recovered Content" in result
    assert any(
        "Attempt 1 failed: Simulated Timeout" in record.message
        for record in caplog.records
    )

    class DummyContext:
        def __init__(self):
            self.new_page_called = False

        async def new_page(self):
            self.new_page_called = True
            return DummyPage()

    class DummyBrowser:
        def __init__(self):
            self.new_context_kwargs = None

        async def new_context(self, **kwargs):
            self.new_context_kwargs = kwargs
            return DummyContext()

        async def close(self):
            return

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())

    # Initialize the loader with a non-empty storage_state value.
    loader = ChromiumLoader(
        ["http://example.com"], backend="playwright", storage_state="dummy_state"
    )

    # Call ascrape_playwright and capture its result.
    result = await loader.ascrape_playwright("http://example.com")

    # To verify that ignore_https_errors was passed into new_context,
    # simulate a separate launch to inspect the new_context_kwargs.
    browser_instance = await DummyPW.chromium.launch(
        headless=loader.headless, proxy=loader.proxy
    )
    await browser_instance.new_context(
        storage_state=loader.storage_state, ignore_https_errors=True
    )
    kwargs = browser_instance.new_context_kwargs

    assert kwargs is not None
    assert kwargs.get("ignore_https_errors") is True
    assert kwargs.get("storage_state") == "dummy_state"
    assert "<html>Ignore HTTPS errors Test</html>" in result


@pytest.mark.asyncio
async def test_ascrape_with_js_support_context_error_cleanup(monkeypatch):
    """Test that ascrape_with_js_support calls browser.close() even if new_context fails."""
    close_called = {"called": False}

    class DummyBrowser:
        async def new_context(self, **kwargs):
            # Force an exception during context creation
            raise Exception("Context error")

        async def close(self):
            close_called["called"] = True

    class DummyPW:
        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return

        class chromium:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

        class firefox:
            @staticmethod
            async def launch(headless, proxy, **kwargs):
                return DummyBrowser()

    monkeypatch.setattr("playwright.async_api.async_playwright", lambda: DummyPW())
    loader = ChromiumLoader(
        ["http://example.com"],
        backend="playwright",
        requires_js_support=True,
        retry_limit=1,
        timeout=1,
    )
    with pytest.raises(RuntimeError, match="Failed to scrape after 1 attempts"):
        await loader.ascrape_with_js_support("http://example.com")
    assert close_called["called"] is True


@pytest.mark.asyncio
async def test_lazy_load_with_none_urls(monkeypatch):
    """Test that lazy_load raises TypeError when urls is None."""
    loader = ChromiumLoader(None, backend="playwright")
    with pytest.raises(TypeError):
        list(loader.lazy_load())


@pytest.mark.asyncio
def test_lazy_load_sequential_timing(monkeypatch):
    """Test that lazy_load runs scraping sequentially rather than concurrently."""
    urls = ["http://example.com/1", "http://example.com/2", "http://example.com/3"]
    loader = ChromiumLoader(urls, backend="playwright", requires_js_support=False)

    async def dummy_scraper_with_delay(url, browser_name="chromium"):
        await asyncio.sleep(0.5)
        return f"<html>Delayed content for {url}</html>"

    monkeypatch.setattr(loader, "ascrape_playwright", dummy_scraper_with_delay)
    start = time.monotonic()
    docs = list(loader.lazy_load())
    elapsed = time.monotonic() - start
    # At least 0.5 seconds per URL should be observed.
    assert elapsed >= 1.5, (
        f"Sequential lazy_load took too little time: {elapsed:.2f} seconds"
    )
    for doc, url in zip(docs, urls):
        assert f"Delayed content for {url}" in doc.page_content
        assert doc.metadata["source"] == url


@pytest.mark.asyncio
def test_lazy_load_with_tuple_urls(monkeypatch):
    """Test that lazy_load yields Document objects correctly when urls is provided as a tuple."""
    urls = ("http://example.com", "http://test.com")
    loader = ChromiumLoader(urls, backend="playwright", requires_js_support=False)

    async def dummy_scraper(url, browser_name="chromium"):
        return f"<html>Tuple content for {url}</html>"

    monkeypatch.setattr(loader, "ascrape_playwright", dummy_scraper)
    docs = list(loader.lazy_load())
    assert len(docs) == 2
    for doc, url in zip(docs, urls):
        assert f"Tuple content for {url}" in doc.page_content
        assert doc.metadata["source"] == url



"""
ScriptCreatorMultiGraph Module
"""

from copy import deepcopy
from typing import List, Optional, Type

from pydantic import BaseModel

from ..nodes import GraphIteratorNode, MergeGeneratedScriptsNode
from ..utils.copy import safe_deepcopy
from .abstract_graph import AbstractGraph
from .base_graph import BaseGraph
from .script_creator_graph import ScriptCreatorGraph


class ScriptCreatorMultiGraph(AbstractGraph):
    """
    ScriptCreatorMultiGraph is a scraping pipeline that scrapes a list
    of URLs generating web scraping scripts.
    It only requires a user prompt and a list of URLs.
    Attributes:
        prompt (str): The user prompt to search the internet.
        llm_model (dict): The configuration for the language model.
        embedder_model (dict): The configuration for the embedder model.
        headless (bool): A flag to run the browser in headless mode.
        verbose (bool): A flag to display the execution information.
        model_token (int): The token limit for the language model.
    Args:
        prompt (str): The user prompt to search the internet.
        source (List[str]): The source of the graph.
        config (dict): Configuration parameters for the graph.
        schema (Optional[BaseModel]): The schema for the graph output.
    Example:
        >>> script_graph = ScriptCreatorMultiGraph(
        ...     "What is Chioggia famous for?",
        ...     source=[],
        ...     config={"llm": {"model": "openai/gpt-3.5-turbo"}}
        ...     schema={}
        ... )
        >>> result = script_graph.run()
    """

    def __init__(
        self,
        prompt: str,
        source: List[str],
        config: dict,
        schema: Optional[Type[BaseModel]] = None,
    ):
        self.copy_config = safe_deepcopy(config)
        self.copy_schema = deepcopy(schema)
        super().__init__(prompt, config, source, schema)

    def _create_graph(self) -> BaseGraph:
        """
        Creates the graph of nodes representing the workflow for web scraping and searching.
        Returns:
            BaseGraph: A graph instance representing the web scraping and searching workflow.
        """

        graph_iterator_node = GraphIteratorNode(
            input="user_prompt & urls",
            output=["scripts"],
            node_config={
                "graph_instance": ScriptCreatorGraph,
                "scraper_config": self.copy_config,
            },
            schema=self.copy_schema,
        )

        merge_scripts_node = MergeGeneratedScriptsNode(
            input="user_prompt & scripts",
            output=["merged_script"],
            node_config={"llm_model": self.llm_model, "schema": self.schema},
        )

        return BaseGraph(
            nodes=[
                graph_iterator_node,
                merge_scripts_node,
            ],
            edges=[
                (graph_iterator_node, merge_scripts_node),
            ],
            entry_point=graph_iterator_node,
            graph_name=self.__class__.__name__,
        )

    def run(self) -> str:
        """
        Executes the web scraping and searching process.
        Returns:
            str: The answer to the prompt.
        """

        inputs = {"user_prompt": self.prompt, "urls": self.source}
        self.final_state, self.execution_info = self.graph.execute(inputs)
        return self.final_state.get("merged_script", "Failed to generate the script.")



"""
Prettify the execution information of the graph.
"""

from typing import Union


def prettify_exec_info(
    complete_result: list[dict], as_string: bool = True
) -> Union[str, list[dict]]:
    """
    Formats the execution information of a graph showing node statistics.

    Args:
        complete_result (list[dict]): The execution information containing node statistics.
        as_string (bool, optional): If True, returns a formatted string table.
                                  If False, returns the original list. Defaults to True.

    Returns:
        Union[str, list[dict]]: A formatted string table if as_string=True,
        otherwise the original list of dictionaries.
    """
    if not as_string:
        return complete_result

    if not complete_result:
        return "Empty result"

    # Format the table
    lines = []
    lines.append("Node Statistics:")
    lines.append("-" * 100)
    lines.append(
        f"{'Node':<20} {'Tokens':<10} {'Prompt':<10} {'Compl.':<10} {'Requests':<10} {'Cost ($)':<10} {'Time (s)':<10}"
    )
    lines.append("-" * 100)

    for item in complete_result:
        node = item["node_name"]
        tokens = item["total_tokens"]
        prompt = item["prompt_tokens"]
        completion = item["completion_tokens"]
        requests = item["successful_requests"]
        cost = f"{item['total_cost_USD']:.4f}"
        time = f"{item['exec_time']:.2f}"

        lines.append(
            f"{node:<20} {tokens:<10} {prompt:<10} {completion:<10} {requests:<10} {cost:<10} {time:<10}"
        )

    return "\n".join(lines)



"""
ScreenshotScraperGraph Module
"""

from typing import Optional, Type

from pydantic import BaseModel

from ..nodes import FetchScreenNode, GenerateAnswerFromImageNode
from .abstract_graph import AbstractGraph
from .base_graph import BaseGraph


class ScreenshotScraperGraph(AbstractGraph):
    """
    A graph instance representing the web scraping workflow for images.

    Attributes:
        prompt (str): The input text to be scraped.
        config (dict): Configuration parameters for the graph.
        source (str): The source URL or image link to scrape from.

    Methods:
        __init__(prompt: str, source: str, config: dict, schema: Optional[Type[BaseModel]] = None)
            Initializes the ScreenshotScraperGraph instance with the given prompt,
            source, and configuration parameters.

        _create_graph()
            Creates a graph of nodes representing the web scraping workflow for images.

        run()
            Executes the scraping process and returns the answer to the prompt.
    """

    def __init__(
        self,
        prompt: str,
        source: str,
        config: dict,
        schema: Optional[Type[BaseModel]] = None,
    ):
        super().__init__(prompt, config, source, schema)

    def _create_graph(self) -> BaseGraph:
        """
        Creates the graph of nodes representing the workflow for web scraping with images.

        Returns:
            BaseGraph: A graph instance representing the web scraping workflow for images.
        """
        fetch_screen_node = FetchScreenNode(
            input="url", output=["screenshots"], node_config={"link": self.source}
        )

        generate_answer_from_image_node = GenerateAnswerFromImageNode(
            input="screenshots", output=["answer"], node_config={"config": self.config}
        )

        return BaseGraph(
            nodes=[
                fetch_screen_node,
                generate_answer_from_image_node,
            ],
            edges=[
                (fetch_screen_node, generate_answer_from_image_node),
            ],
            entry_point=fetch_screen_node,
            graph_name=self.__class__.__name__,
        )

    def run(self) -> str:
        """
        Executes the scraping process and returns the answer to the prompt.

        Returns:
            str: The answer to the prompt.
        """

        inputs = {"user_prompt": self.prompt}
        self.final_state, self.execution_info = self.graph.execute(inputs)

        return self.final_state.get("answer", "No answer found.")



"""
Basic example of scraping pipeline using SmartScraper with schema
"""

import json

from pydantic import BaseModel, Field

from scrapegraphai.graphs import SmartScraperGraph
from scrapegraphai.utils import prettify_exec_info


# ************************************************
# Define the configuration for the graph
# ************************************************
class Project(BaseModel):
    title: str = Field(description="The title of the project")
    description: str = Field(description="The description of the project")


class Projects(BaseModel):
    projects: list[Project]


graph_config = {
    "llm": {"model": "ollama/llama3.2", "temperature": 0, "model_tokens": 4096},
    "verbose": True,
    "headless": False,
}

# ************************************************
# Create the SmartScraperGraph instance and run it
# ************************************************

smart_scraper_graph = SmartScraperGraph(
    prompt="List me all the projects with their description",
    source="https://perinim.github.io/projects/",
    schema=Projects,
    config=graph_config,
)

result = smart_scraper_graph.run()
print(json.dumps(result, indent=4))

# ************************************************
# Get graph execution info
# ************************************************

graph_exec_info = smart_scraper_graph.get_execution_info()
print(prettify_exec_info(graph_exec_info))



"""
GenerateAnswerNode Module
"""

import json
import time
from typing import List, Optional

from langchain.prompts import PromptTemplate
from langchain_aws import ChatBedrock
from langchain_community.chat_models import ChatOllama
from langchain_core.output_parsers import JsonOutputParser
from langchain_core.runnables import RunnableParallel
from langchain_openai import ChatOpenAI
from requests.exceptions import Timeout
from tqdm import tqdm

from ..prompts import (
    TEMPLATE_CHUNKS,
    TEMPLATE_CHUNKS_MD,
    TEMPLATE_MERGE,
    TEMPLATE_MERGE_MD,
    TEMPLATE_NO_CHUNKS,
    TEMPLATE_NO_CHUNKS_MD,
)
from ..utils.output_parser import get_pydantic_output_parser
from .base_node import BaseNode


class GenerateAnswerNode(BaseNode):
    """
    Initializes the GenerateAnswerNode class.

    Args:
        input (str): The input data type for the node.
        output (List[str]): The output data type(s) for the node.
        node_config (Optional[dict]): Configuration dictionary for the node,
        which includes the LLM model, verbosity, schema, and other settings.
        Defaults to None.
        node_name (str): The name of the node. Defaults to "GenerateAnswer".

    Attributes:
        llm_model: The language model specified in the node configuration.
        verbose (bool): Whether verbose mode is enabled.
        force (bool): Whether to force certain behaviors, overriding defaults.
        script_creator (bool): Whether the node is in script creation mode.
        is_md_scraper (bool): Whether the node is scraping markdown data.
        additional_info (Optional[str]): Any additional information to be
        included in the prompt templates.
    """

    def __init__(
        self,
        input: str,
        output: List[str],
        node_config: Optional[dict] = None,
        node_name: str = "GenerateAnswer",
    ):
        super().__init__(node_name, "node", input, output, 2, node_config)
        self.llm_model = node_config["llm_model"]

        if isinstance(node_config["llm_model"], ChatOllama):
            if node_config.get("schema", None) is None:
                self.llm_model.format = "json"
            else:
                self.llm_model.format = self.node_config["schema"].model_json_schema()

        self.verbose = node_config.get("verbose", False)
        self.force = node_config.get("force", False)
        self.script_creator = node_config.get("script_creator", False)
        self.is_md_scraper = node_config.get("is_md_scraper", False)
        self.additional_info = node_config.get("additional_info")
        self.timeout = node_config.get("timeout", 480)

    def invoke_with_timeout(self, chain, inputs, timeout):
        """Helper method to invoke chain with timeout"""
        try:
            start_time = time.time()
            response = chain.invoke(inputs)
            if time.time() - start_time > timeout:
                raise Timeout(f"Response took longer than {timeout} seconds")
            return response
        except Timeout as e:
            self.logger.error(f"Timeout error: {str(e)}")
            raise
        except Exception as e:
            self.logger.error(f"Error during chain execution: {str(e)}")
            raise

    def process(self, state: dict) -> dict:
        """Process the input state and generate an answer."""
        user_prompt = state.get("user_prompt")
        # Check for content in different possible state keys
        content = (
            state.get("relevant_chunks")
            or state.get("parsed_doc")
            or state.get("doc")
            or state.get("content")
        )

        if not content:
            raise ValueError("No content found in state to generate answer from")

        if not user_prompt:
            raise ValueError("No user prompt found in state")

        # Create the chain input with both content and question keys
        chain_input = {"content": content, "question": user_prompt}

        try:
            response = self.invoke_with_timeout(self.chain, chain_input, self.timeout)
            state.update({self.output[0]: response})
            return state
        except Exception as e:
            self.logger.error(f"Error in GenerateAnswerNode: {str(e)}")
            raise

    def execute(self, state: dict) -> dict:
        """
        Executes the GenerateAnswerNode.

        Args:
            state (dict): The current state of the graph. The input keys will be used
                          to fetch the correct data from the state.

        Returns:
            dict: The updated state with the output key containing the generated answer.
        """
        self.logger.info(f"--- Executing {self.node_name} Node ---")

        input_keys = self.get_input_keys(state)
        input_data = [state[key] for key in input_keys]
        user_prompt = input_data[0]
        doc = input_data[1]

        if self.node_config.get("schema", None) is not None:
            if isinstance(self.llm_model, ChatOpenAI):
                output_parser = get_pydantic_output_parser(self.node_config["schema"])
                format_instructions = output_parser.get_format_instructions()
            else:
                if not isinstance(self.llm_model, ChatBedrock):
                    output_parser = get_pydantic_output_parser(
                        self.node_config["schema"]
                    )
                    format_instructions = output_parser.get_format_instructions()
                else:
                    output_parser = None
                    format_instructions = ""
        else:
            if not isinstance(self.llm_model, ChatBedrock):
                output_parser = JsonOutputParser()
                format_instructions = (
                    "You must respond with a JSON object. Your response should be formatted as a valid JSON "
                    "with a 'content' field containing your analysis. For example:\n"
                    '{{"content": "your analysis here"}}'
                )
            else:
                output_parser = None
                format_instructions = ""

        if (
            not self.script_creator
            or self.force
            and not self.script_creator
            or self.is_md_scraper
        ):
            template_no_chunks_prompt = TEMPLATE_NO_CHUNKS_MD
            template_chunks_prompt = TEMPLATE_CHUNKS_MD
            template_merge_prompt = TEMPLATE_MERGE_MD
        else:
            template_no_chunks_prompt = TEMPLATE_NO_CHUNKS
            template_chunks_prompt = TEMPLATE_CHUNKS
            template_merge_prompt = TEMPLATE_MERGE

        if self.additional_info is not None:
            template_no_chunks_prompt = self.additional_info + template_no_chunks_prompt
            template_chunks_prompt = self.additional_info + template_chunks_prompt
            template_merge_prompt = self.additional_info + template_merge_prompt

        if len(doc) == 1:
            prompt = PromptTemplate(
                template=template_no_chunks_prompt,
                input_variables=["question"],
                partial_variables={
                    "context": doc,
                    "format_instructions": format_instructions,
                },
            )
            chain = prompt | self.llm_model
            if output_parser:
                chain = chain | output_parser

            try:
                answer = self.invoke_with_timeout(
                    chain, {"question": user_prompt}, self.timeout
                )
            except (Timeout, json.JSONDecodeError) as e:
                error_msg = (
                    "Response timeout exceeded"
                    if isinstance(e, Timeout)
                    else "Invalid JSON response format"
                )
                state.update(
                    {self.output[0]: {"error": error_msg, "raw_response": str(e)}}
                )
                return state

            state.update({self.output[0]: answer})
            return state

        chains_dict = {}
        for i, chunk in enumerate(
            tqdm(doc, desc="Processing chunks", disable=not self.verbose)
        ):
            prompt = PromptTemplate(
                template=template_chunks_prompt,
                input_variables=["question"],
                partial_variables={
                    "context": chunk,
                    "chunk_id": i + 1,
                    "format_instructions": format_instructions,
                },
            )
            chain_name = f"chunk{i + 1}"
            chains_dict[chain_name] = prompt | self.llm_model
            if output_parser:
                chains_dict[chain_name] = chains_dict[chain_name] | output_parser

        async_runner = RunnableParallel(**chains_dict)
        try:
            batch_results = self.invoke_with_timeout(
                async_runner, {"question": user_prompt}, self.timeout
            )
        except (Timeout, json.JSONDecodeError) as e:
            error_msg = (
                "Response timeout exceeded during chunk processing"
                if isinstance(e, Timeout)
                else "Invalid JSON response format in chunk processing"
            )
            state.update({self.output[0]: {"error": error_msg, "raw_response": str(e)}})
            return state

        merge_prompt = PromptTemplate(
            template=template_merge_prompt,
            input_variables=["context", "question"],
            partial_variables={"format_instructions": format_instructions},
        )

        merge_chain = merge_prompt | self.llm_model
        if output_parser:
            merge_chain = merge_chain | output_parser
        try:
            answer = self.invoke_with_timeout(
                merge_chain,
                {"context": batch_results, "question": user_prompt},
                self.timeout,
            )
        except (Timeout, json.JSONDecodeError) as e:
            error_msg = (
                "Response timeout exceeded during merge"
                if isinstance(e, Timeout)
                else "Invalid JSON response format during merge"
            )
            state.update({self.output[0]: {"error": error_msg, "raw_response": str(e)}})
            return state

        state.update({self.output[0]: answer})
        return state



"""
__init__.py for the prompts folder
"""

from .generate_answer_node_csv_prompts import (
    TEMPLATE_CHUKS_CSV,
    TEMPLATE_MERGE_CSV,
    TEMPLATE_NO_CHUKS_CSV,
)
from .generate_answer_node_omni_prompts import (
    TEMPLATE_CHUNKS_OMNI,
    TEMPLATE_MERGE_OMNI,
    TEMPLATE_NO_CHUNKS_OMNI,
)
from .generate_answer_node_pdf_prompts import (
    TEMPLATE_CHUNKS_PDF,
    TEMPLATE_MERGE_PDF,
    TEMPLATE_NO_CHUNKS_PDF,
)
from .generate_answer_node_prompts import (
    REGEN_ADDITIONAL_INFO,
    TEMPLATE_CHUNKS,
    TEMPLATE_CHUNKS_MD,
    TEMPLATE_MERGE,
    TEMPLATE_MERGE_MD,
    TEMPLATE_NO_CHUNKS,
    TEMPLATE_NO_CHUNKS_MD,
)
from .generate_code_node_prompts import (
    TEMPLATE_EXECUTION_ANALYSIS,
    TEMPLATE_EXECUTION_CODE_GENERATION,
    TEMPLATE_INIT_CODE_GENERATION,
    TEMPLATE_SEMANTIC_ANALYSIS,
    TEMPLATE_SEMANTIC_CODE_GENERATION,
    TEMPLATE_SEMANTIC_COMPARISON,
    TEMPLATE_SYNTAX_ANALYSIS,
    TEMPLATE_SYNTAX_CODE_GENERATION,
    TEMPLATE_VALIDATION_ANALYSIS,
    TEMPLATE_VALIDATION_CODE_GENERATION,
)
from .get_probable_tags_node_prompts import TEMPLATE_GET_PROBABLE_TAGS
from .html_analyzer_node_prompts import (
    TEMPLATE_HTML_ANALYSIS,
    TEMPLATE_HTML_ANALYSIS_WITH_CONTEXT,
)
from .merge_answer_node_prompts import TEMPLATE_COMBINED
from .merge_generated_scripts_prompts import TEMPLATE_MERGE_SCRIPTS_PROMPT
from .prompt_refiner_node_prompts import TEMPLATE_REFINER, TEMPLATE_REFINER_WITH_CONTEXT
from .reasoning_node_prompts import TEMPLATE_REASONING, TEMPLATE_REASONING_WITH_CONTEXT
from .robots_node_prompts import TEMPLATE_ROBOT
from .search_internet_node_prompts import TEMPLATE_SEARCH_INTERNET
from .search_link_node_prompts import TEMPLATE_RELEVANT_LINKS
from .search_node_with_context_prompts import (
    TEMPLATE_SEARCH_WITH_CONTEXT_CHUNKS,
    TEMPLATE_SEARCH_WITH_CONTEXT_NO_CHUNKS,
)

__all__ = [
    # CSV Answer Generation Templates
    "TEMPLATE_CHUKS_CSV",
    "TEMPLATE_MERGE_CSV",
    "TEMPLATE_NO_CHUKS_CSV",
    # Omni Answer Generation Templates
    "TEMPLATE_CHUNKS_OMNI",
    "TEMPLATE_MERGE_OMNI",
    "TEMPLATE_NO_CHUNKS_OMNI",
    # PDF Answer Generation Templates
    "TEMPLATE_CHUNKS_PDF",
    "TEMPLATE_MERGE_PDF",
    "TEMPLATE_NO_CHUNKS_PDF",
    # General Answer Generation Templates
    "REGEN_ADDITIONAL_INFO",
    "TEMPLATE_CHUNKS",
    "TEMPLATE_CHUNKS_MD",
    "TEMPLATE_MERGE",
    "TEMPLATE_MERGE_MD",
    "TEMPLATE_NO_CHUNKS",
    "TEMPLATE_NO_CHUNKS_MD",
    # Code Generation and Analysis Templates
    "TEMPLATE_EXECUTION_ANALYSIS",
    "TEMPLATE_EXECUTION_CODE_GENERATION",
    "TEMPLATE_INIT_CODE_GENERATION",
    "TEMPLATE_SEMANTIC_ANALYSIS",
    "TEMPLATE_SEMANTIC_CODE_GENERATION",
    "TEMPLATE_SEMANTIC_COMPARISON",
    "TEMPLATE_SYNTAX_ANALYSIS",
    "TEMPLATE_SYNTAX_CODE_GENERATION",
    "TEMPLATE_VALIDATION_ANALYSIS",
    "TEMPLATE_VALIDATION_CODE_GENERATION",
    # HTML and Tag Analysis Templates
    "TEMPLATE_GET_PROBABLE_TAGS",
    "TEMPLATE_HTML_ANALYSIS",
    "TEMPLATE_HTML_ANALYSIS_WITH_CONTEXT",
    # Merging and Combining Templates
    "TEMPLATE_COMBINED",
    "TEMPLATE_MERGE_SCRIPTS_PROMPT",
    # Search and Context Templates
    "TEMPLATE_SEARCH_INTERNET",
    "TEMPLATE_RELEVANT_LINKS",
    "TEMPLATE_SEARCH_WITH_CONTEXT_CHUNKS",
    "TEMPLATE_SEARCH_WITH_CONTEXT_NO_CHUNKS",
    # Reasoning and Refinement Templates
    "TEMPLATE_REFINER",
    "TEMPLATE_REFINER_WITH_CONTEXT",
    "TEMPLATE_REASONING",
    "TEMPLATE_REASONING_WITH_CONTEXT",
    # Robot Templates
    "TEMPLATE_ROBOT",
]



"""
Module for filtering irrelevant links
"""

filter_dict = {
    "diff_domain_filter": True,
    "img_exts": [".jpg", ".jpeg", ".png", ".gif", ".bmp", ".svg", ".webp", ".ico"],
    "lang_indicators": ["lang=", "/fr", "/pt", "/es", "/de", "/jp", "/it"],
    "irrelevant_keywords": [
        "/login",
        "/signup",
        "/register",
        "/contact",
        "facebook.com",
        "twitter.com",
        "linkedin.com",
        "instagram.com",
        ".js",
        ".css",
    ],
}



from unittest.mock import MagicMock, patch

import pytest

from scrapegraphai.graphs.search_graph import SearchGraph


class TestSearchGraph:
    """Test class for SearchGraph"""

    @pytest.mark.parametrize(
        "urls",
        [["https://example.com", "https://test.com"], [], ["https://single-url.com"]],
    )
    @patch("scrapegraphai.graphs.search_graph.BaseGraph")
    @patch("scrapegraphai.graphs.abstract_graph.AbstractGraph._create_llm")
    def test_get_considered_urls(self, mock_create_llm, mock_base_graph, urls):
        """
        Test that get_considered_urls returns the correct list of URLs
        considered during the search process.
        """
        # Arrange
        prompt = "Test prompt"
        config = {"llm": {"model": "test-model"}}

        # Mock the _create_llm method to return a MagicMock
        mock_create_llm.return_value = MagicMock()

        # Mock the execute method to set the final_state
        mock_base_graph.return_value.execute.return_value = ({"urls": urls}, {})

        # Act
        search_graph = SearchGraph(prompt, config)
        search_graph.run()

        # Assert
        assert search_graph.get_considered_urls() == urls

    @patch("scrapegraphai.graphs.search_graph.BaseGraph")
    @patch("scrapegraphai.graphs.abstract_graph.AbstractGraph._create_llm")
    def test_run_no_answer_found(self, mock_create_llm, mock_base_graph):
        """
        Test that the run() method returns "No answer found." when the final state
        doesn't contain an "answer" key.
        """
        # Arrange
        prompt = "Test prompt"
        config = {"llm": {"model": "test-model"}}

        # Mock the _create_llm method to return a MagicMock
        mock_create_llm.return_value = MagicMock()

        # Mock the execute method to set the final_state without an "answer" key
        mock_base_graph.return_value.execute.return_value = ({"urls": []}, {})

        # Act
        search_graph = SearchGraph(prompt, config)
        result = search_graph.run()

        # Assert
        assert result == "No answer found."

    @patch("scrapegraphai.graphs.search_graph.SearchInternetNode")
    @patch("scrapegraphai.graphs.search_graph.GraphIteratorNode")
    @patch("scrapegraphai.graphs.search_graph.MergeAnswersNode")
    @patch("scrapegraphai.graphs.search_graph.BaseGraph")
    @patch("scrapegraphai.graphs.abstract_graph.AbstractGraph._create_llm")
    def test_max_results_config(
        self,
        mock_create_llm,
        mock_base_graph,
        mock_merge_answers,
        mock_graph_iterator,
        mock_search_internet,
    ):
        """
        Test that the max_results parameter from the config is correctly passed to the SearchInternetNode.
        """
        # Arrange
        prompt = "Test prompt"
        max_results = 5
        config = {"llm": {"model": "test-model"}, "max_results": max_results}

        # Act
        search_graph = SearchGraph(prompt, config)

        # Assert
        mock_search_internet.assert_called_once()
        call_args = mock_search_internet.call_args
        assert call_args.kwargs["node_config"]["max_results"] == max_results

    @patch("scrapegraphai.graphs.search_graph.SearchInternetNode")
    @patch("scrapegraphai.graphs.search_graph.GraphIteratorNode")
    @patch("scrapegraphai.graphs.search_graph.MergeAnswersNode")
    @patch("scrapegraphai.graphs.search_graph.BaseGraph")
    @patch("scrapegraphai.graphs.abstract_graph.AbstractGraph._create_llm")
    def test_custom_search_engine_config(
        self,
        mock_create_llm,
        mock_base_graph,
        mock_merge_answers,
        mock_graph_iterator,
        mock_search_internet,
    ):
        """
        Test that the custom search_engine parameter from the config is correctly passed to the SearchInternetNode.
        """
        # Arrange
        prompt = "Test prompt"
        custom_search_engine = "custom_engine"
        config = {"llm": {"model": "test-model"}, "search_engine": custom_search_engine}

        # Act
        search_graph = SearchGraph(prompt, config)

        # Assert
        mock_search_internet.assert_called_once()
        call_args = mock_search_internet.call_args
        assert call_args.kwargs["node_config"]["search_engine"] == custom_search_engine



from copy import deepcopy

import pytest

from scrapegraphai.graphs.csv_scraper_multi_graph import CSVScraperMultiGraph

# Monkey-patch _create_llm to avoid unsupported provider error during tests
CSVScraperMultiGraph._create_llm = lambda self, llm_config: llm_config


# Dummy graph classes to simulate behavior during tests
class DummyGraph:
    """Dummy graph that returns a predefined answer."""

    def __init__(self, answer):
        self.answer = answer

    def execute(self, inputs):
        # Returns a tuple of (final_state, execution_info)
        return ({"answer": self.answer}, {})


class DummyGraphNoAnswer:
    """Dummy graph that simulates absence of answer in final_state."""

    def execute(self, inputs):
        # Returns an empty final_state
        return ({}, {})


class DummyBaseGraph:
    """Dummy BaseGraph to test _create_graph method without side effects."""

    def __init__(self, nodes, edges, entry_point, graph_name):
        self.nodes = nodes
        self.edges = edges
        self.entry_point = entry_point
        self.graph_name = graph_name

    config = {
        "llm": {"model": "dummy_model", "model_provider": "dummy_provider"},
        "key": "value",
    }
    """Test that CSVScraperMultiGraph.run returns the expected answer when provided by the graph."""
    prompt = "Test prompt"
    source = ["url1", "url2"]

    # Instantiate the graph
    multi_graph = CSVScraperMultiGraph(prompt, source, config)

    # Override the graph attribute with a dummy graph returning an expected answer
    multi_graph.graph = DummyGraph("expected answer")

    result = multi_graph.run()
    assert result == "expected answer"


def test_run_no_answer():
    """Test that CSVScraperMultiGraph.run returns a fallback message when no answer is provided."""
    prompt = "Another test prompt"
    source = ["url3"]
    config = {
        "llm": {"model": "dummy_model", "model_provider": "dummy_provider"},
        "another_key": "another_value",
    }

    multi_graph = CSVScraperMultiGraph(prompt, source, config)
    multi_graph.graph = DummyGraphNoAnswer()

    result = multi_graph.run()
    assert result == "No answer found."


def test_create_graph_structure(monkeypatch):
    """Test that _create_graph constructs a graph with the expected structure."""
    prompt = "Structure test"
    source = ["url4"]
    config = {
        "llm": {"model": "dummy_model", "model_provider": "dummy_provider"},
        "struct_key": "struct_value",
    }

    multi_graph = CSVScraperMultiGraph(prompt, source, config)

    # Monkey-patch the _create_graph method to avoid dependencies on external nodes
    monkeypatch.setattr(
        multi_graph,
        "_create_graph",
        lambda: DummyBaseGraph(
            nodes=["graph_iterator_node", "merge_answers_node"],
            edges=[("graph_iterator_node", "merge_answers_node")],
            entry_point="graph_iterator_node",
            graph_name=multi_graph.__class__.__name__,
        ),
    )

    graph = multi_graph._create_graph()
    assert graph.graph_name == "CSVScraperMultiGraph"
    assert len(graph.nodes) == 2
    assert len(graph.edges) == 1


def test_config_deepcopy():
    """Test that the configuration dictionary is deep-copied.
    Modifying the original config after instantiation should not affect the multi_graph copy.
    """
    config = {
        "llm": {"model": "dummy_model", "provider": "provider1"},
        "nested": {"a": [1, 2]},
    }
    original_config = deepcopy(config)
    multi_graph = CSVScraperMultiGraph("Deep copy test", ["url_deep"], config)
    # Modify the original config after instantiation
    config["nested"]["a"].append(3)
    # The multi_graph.copy_config should remain unchanged.
    assert multi_graph.copy_config["nested"]["a"] == original_config["nested"]["a"]


def test_run_argument_passing():
    """Test that CSVScraperMultiGraph.run passes the correct input arguments
    to the graph's execute method and returns the expected answer."""

    class DummyGraphCapture:
        def __init__(self):
            self.captured_inputs = None

        def execute(self, inputs):
            self.captured_inputs = inputs
            return ({"answer": "captured answer"}, {})

    prompt = "Argument test prompt"
    source = ["url_arg1", "url_arg2"]
    config = {"llm": {"model": "dummy_model", "provider": "dummy_provider"}}

    multi_graph = CSVScraperMultiGraph(prompt, source, config)
    dummy_graph = DummyGraphCapture()
    multi_graph.graph = dummy_graph

    result = multi_graph.run()
    # Check that the dummy graph captured the inputs as expected
    expected_inputs = {"user_prompt": prompt, "jsons": source}
    assert dummy_graph.captured_inputs == expected_inputs
    assert result == "captured answer"


def test_run_with_exception_in_execute():
    """Test that CSVScraperMultiGraph.run propagates exceptions from the graph's execute method."""

    class DummyGraphException:
        def execute(self, inputs):
            raise Exception("Test exception")

    prompt = "Exception test prompt"
    source = ["url_exception"]
    config = {"llm": {"model": "dummy_model", "provider": "dummy_provider"}}

    multi_graph = CSVScraperMultiGraph(prompt, source, config)
    multi_graph.graph = DummyGraphException()

    with pytest.raises(Exception, match="Test exception"):
        multi_graph.run()



"""
OmniSearchGraph Module
"""

from copy import deepcopy
from typing import Optional, Type

from pydantic import BaseModel

from ..nodes import GraphIteratorNode, MergeAnswersNode, SearchInternetNode
from ..utils.copy import safe_deepcopy
from .abstract_graph import AbstractGraph
from .base_graph import BaseGraph
from .omni_scraper_graph import OmniScraperGraph


class OmniSearchGraph(AbstractGraph):
    """
    OmniSearchGraph is a scraping pipeline that searches the internet for answers to a given prompt.
    It only requires a user prompt to search the internet and generate an answer.

    Attributes:
        prompt (str): The user prompt to search the internet.
        llm_model (dict): The configuration for the language model.
        embedder_model (dict): The configuration for the embedder model.
        headless (bool): A flag to run the browser in headless mode.
        verbose (bool): A flag to display the execution information.
        model_token (int): The token limit for the language model.
        max_results (int): The maximum number of results to return.

    Args:
        prompt (str): The user prompt to search the internet.
        config (dict): Configuration parameters for the graph.
        schema (Optional[BaseModel]): The schema for the graph output.

    Example:
        >>> omni_search_graph = OmniSearchGraph(
        ...     "What is Chioggia famous for?",
        ...     {"llm": {"model": "openai/gpt-4o"}}
        ... )
        >>> result = search_graph.run()
    """

    def __init__(
        self, prompt: str, config: dict, schema: Optional[Type[BaseModel]] = None
    ):
        self.max_results = config.get("max_results", 3)

        self.copy_config = safe_deepcopy(config)

        self.copy_schema = deepcopy(schema)

        super().__init__(prompt, config, schema)

    def _create_graph(self) -> BaseGraph:
        """
        Creates the graph of nodes representing the workflow for web scraping and searching.

        Returns:
            BaseGraph: A graph instance representing the web scraping and searching workflow.
        """

        search_internet_node = SearchInternetNode(
            input="user_prompt",
            output=["urls"],
            node_config={
                "llm_model": self.llm_model,
                "max_results": self.max_results,
                "search_engine": self.copy_config.get("search_engine"),
            },
        )
        graph_iterator_node = GraphIteratorNode(
            input="user_prompt & urls",
            output=["results"],
            node_config={
                "graph_instance": OmniScraperGraph,
                "scraper_config": self.copy_config,
            },
            schema=self.copy_schema,
        )

        merge_answers_node = MergeAnswersNode(
            input="user_prompt & results",
            output=["answer"],
            node_config={"llm_model": self.llm_model, "schema": self.copy_schema},
        )

        return BaseGraph(
            nodes=[search_internet_node, graph_iterator_node, merge_answers_node],
            edges=[
                (search_internet_node, graph_iterator_node),
                (graph_iterator_node, merge_answers_node),
            ],
            entry_point=search_internet_node,
            graph_name=self.__class__.__name__,
        )

    def run(self) -> str:
        """
        Executes the web scraping and searching process.

        Returns:
            str: The answer to the prompt.
        """

        inputs = {"user_prompt": self.prompt}
        self.final_state, self.execution_info = self.graph.execute(inputs)

        return self.final_state.get("answer", "No answer found.")



"""
Tokenization utilities for OpenAI models
"""

import tiktoken

from ..logging import get_logger


def num_tokens_openai(text: str) -> int:
    """
    Estimate the number of tokens in a given text using OpenAI's tokenization method,
    adjusted for different OpenAI models.

    Args:
        text (str): The text to be tokenized and counted.

    Returns:
        int: The number of tokens in the text.
    """

    logger = get_logger()

    logger.debug(f"Counting tokens for text of {len(text)} characters")

    encoding = tiktoken.encoding_for_model("gpt-4o")

    num_tokens = len(encoding.encode(text))
    return num_tokens



"""
Basic example of scraping pipeline using SmartScraper
"""

import json

from scrapegraphai.graphs import SmartScraperMultiLiteGraph
from scrapegraphai.utils import prettify_exec_info

# ************************************************
# Define the configuration for the graph
# ************************************************

graph_config = {
    "llm": {
        "model": "ollama/llama3.1",
        "temperature": 0,
        "base_url": "http://localhost:11434",  # set ollama URL arbitrarily
    },
    "verbose": True,
    "headless": False,
}

# ************************************************
# Create the SmartScraperGraph instance and run it
# ************************************************

smart_scraper_multi_lite_graph = SmartScraperMultiLiteGraph(
    prompt="Who is ?",
    source=["https://perinim.github.io/", "https://perinim.github.io/cv/"],
    config=graph_config,
)

result = smart_scraper_multi_lite_graph.run()
print(json.dumps(result, indent=4))

# ************************************************
# Get graph execution info
# ************************************************

graph_exec_info = smart_scraper_multi_lite_graph.get_execution_info()
print(prettify_exec_info(graph_exec_info))



import pytest
from pydantic import BaseModel

from scrapegraphai.graphs.script_creator_graph import ScriptCreatorGraph
from scrapegraphai.graphs.script_creator_multi_graph import (
    BaseGraph,
    ScriptCreatorMultiGraph,
)


@pytest.fixture(autouse=True)
def set_api_key_env(monkeypatch):
    monkeypatch.setenv("OPENAI_API_KEY", "dummy")


# Dummy classes to simulate behavior for testing
class DummyGraph:
    def __init__(self, final_state, execution_info):
        self.final_state = final_state
        self.execution_info = execution_info

    def execute(self, inputs):
        return self.final_state, self.execution_info


class DummySchema(BaseModel):
    field: str = "dummy"


class TestScriptCreatorMultiGraph:
    """Tests for ScriptCreatorMultiGraph."""

    def test_run_success(self):
        """Test run() returns the merged script when execution is successful."""
        prompt = "Test prompt"
        source = ["http://example.com"]
        config = {"llm": {"model": "openai/test-model"}}
        schema = DummySchema
        instance = ScriptCreatorMultiGraph(prompt, source, config, schema)
        # Set necessary attributes that are expected by _create_graph() and the run() method.
        instance.llm_model = {"model": "openai/test-model"}
        instance.schema = {"type": "dummy"}
        # Replace the graph with a dummy graph that simulates successful execution.
        dummy_final_state = {"merged_script": "print('Hello World')"}
        dummy_execution_info = {"info": "dummy"}
        instance.graph = DummyGraph(dummy_final_state, dummy_execution_info)
        result = instance.run()
        assert result == "print('Hello World')"

    def test_run_failure(self):
        """Test run() returns failure message when merged_script is missing."""
        prompt = "Test prompt"
        source = ["http://example.com"]
        config = {"llm": {"model": "openai/test-model"}}
        schema = DummySchema
        instance = ScriptCreatorMultiGraph(prompt, source, config, schema)
        instance.llm_model = {"model": "openai/test-model"}
        instance.schema = {"type": "dummy"}
        dummy_final_state = {"other_key": "no script"}
        dummy_execution_info = {"info": "dummy"}
        instance.graph = DummyGraph(dummy_final_state, dummy_execution_info)
        result = instance.run()
        assert result == "Failed to generate the script."

    def test_create_graph_structure(self):
        """Test _create_graph() returns a BaseGraph with the correct graph name and structure."""
        prompt = "Test prompt"
        source = []
        config = {"llm": {"model": "openai/test-model"}}
        schema = DummySchema
        instance = ScriptCreatorMultiGraph(prompt, source, config, schema)
        # Manually assign llm_model and schema for node configuration in the graph.
        instance.llm_model = {"model": "openai/test-model"}
        instance.schema = {"type": "dummy"}
        graph = instance._create_graph()
        assert isinstance(graph, BaseGraph)
        assert hasattr(graph, "graph_name")
        assert graph.graph_name == "ScriptCreatorMultiGraph"
        # Check that the graph has two nodes.
        assert len(graph.nodes) == 2
        # Optional: Check that the edges list is correctly formed.
        assert len(graph.edges) == 1

    def test_config_deepcopy(self):
        """Test that the configuration is deep copied during initialization."""
        prompt = "Test prompt"
        source = []
        config = {"llm": {"model": "openai/test-model"}, "other": [1, 2, 3]}
        schema = DummySchema
        instance = ScriptCreatorMultiGraph(prompt, source, config, schema)
        # Modify the original config.
        config["llm"]["model"] = "changed-model"
        config["other"].append(4)
        # Verify that the config copied within instance remains unchanged.
        assert instance.copy_config["llm"]["model"] == "openai/test-model"
        assert instance.copy_config["other"] == [1, 2, 3]

    def test_init_attributes(self):
        """Test that initial attributes are set correctly upon initialization."""
        prompt = "Initialization test"
        source = ["http://init.com"]
        config = {"llm": {"model": "openai/init-model"}, "param": [1, 2]}
        schema = DummySchema
        instance = ScriptCreatorMultiGraph(prompt, source, config, schema)
        # Check that basic attributes are set correctly
        assert instance.prompt == prompt
        assert instance.source == source
        # Check that copy_config is a deep copy and equals the original config
        assert instance.copy_config == {
            "llm": {"model": "openai/init-model"},
            "param": [1, 2],
        }
        # For classes, deepcopy returns the same object, so the copy_schema should equal schema
        assert instance.copy_schema == DummySchema

    def test_run_no_schema(self):
        """Test run() when schema is None."""
        prompt = "No schema prompt"
        source = ["http://noschema.com"]
        config = {"llm": {"model": "openai/no-schema"}}
        instance = ScriptCreatorMultiGraph(prompt, source, config, schema=None)
        instance.llm_model = {"model": "openai/no-schema"}
        instance.schema = None
        dummy_final_state = {"merged_script": "print('No Schema Script')"}
        dummy_execution_info = {"info": "no schema"}
        instance.graph = DummyGraph(dummy_final_state, dummy_execution_info)
        result = instance.run()
        assert result == "print('No Schema Script')"

    def test_create_graph_node_configs(self):
        """Test that _create_graph() sets correct node configurations for its nodes."""
        prompt = "Graph config test"
        source = ["http://graphconfig.com"]
        config = {"llm": {"model": "openai/graph-model"}, "extra": [10]}
        schema = DummySchema
        instance = ScriptCreatorMultiGraph(prompt, source, config, schema)
        # Manually assign llm_model and schema for node configuration
        instance.llm_model = {"model": "openai/graph-model"}
        instance.schema = {"type": "graph-dummy"}
        graph = instance._create_graph()
        # Validate configuration of the first node (GraphIteratorNode)
        node1 = graph.nodes[0]
        assert node1.node_config["graph_instance"] == ScriptCreatorGraph
        assert node1.node_config["scraper_config"] == instance.copy_config
        # Validate configuration of the second node (MergeGeneratedScriptsNode)
        node2 = graph.nodes[1]
        assert node2.node_config["llm_model"] == instance.llm_model
        assert node2.node_config["schema"] == instance.schema

    def test_entry_point_node(self):
        """Test that the graph entry point is the GraphIteratorNode (the first node)."""
        prompt = "Entry point test"
        source = ["http://entrypoint.com"]
        config = {"llm": {"model": "openai/test-model"}}
        schema = DummySchema
        instance = ScriptCreatorMultiGraph(prompt, source, config, schema)
        instance.llm_model = {"model": "openai/test-model"}
        instance.schema = {"type": "dummy"}
        graph = instance._create_graph()
        assert graph.entry_point == graph.nodes[0]

    def test_run_exception(self):
        """Test that run() propagates exceptions raised by graph.execute."""
        prompt = "Exception test"
        source = ["http://exception.com"]
        config = {"llm": {"model": "openai/test-model"}}
        schema = DummySchema
        instance = ScriptCreatorMultiGraph(prompt, source, config, schema)
        instance.llm_model = {"model": "openai/test-model"}
        instance.schema = {"type": "dummy"}

        # Create a dummy graph that raises an exception when execute is called.
        class ExceptionGraph:
            def execute(self, inputs):
                raise ValueError("Testing exception")

        instance.graph = ExceptionGraph()
        with pytest.raises(ValueError, match="Testing exception"):
            instance.run()

    def test_run_with_empty_prompt(self):
        """Test run() method with an empty prompt."""
        prompt = ""
        source = ["http://emptyprompt.com"]
        config = {"llm": {"model": "openai/test-model"}}
        schema = DummySchema
        instance = ScriptCreatorMultiGraph(prompt, source, config, schema)
        instance.llm_model = {"model": "openai/test-model"}
        instance.schema = {"type": "dummy"}
        dummy_final_state = {"merged_script": "print('Empty prompt')"}
        dummy_execution_info = {"info": "empty prompt"}
        instance.graph = DummyGraph(dummy_final_state, dummy_execution_info)
        result = instance.run()
        assert result == "print('Empty prompt')"

    def test_run_called_twice(self):
        """Test that running run() twice returns consistent and updated results."""
        prompt = "Twice test"
        source = ["http://twicetest.com"]
        config = {"llm": {"model": "openai/test-model"}}
        schema = DummySchema
        instance = ScriptCreatorMultiGraph(prompt, source, config, schema)
        instance.llm_model = {"model": "openai/test-model"}
        instance.schema = {"type": "dummy"}
        dummy_final_state = {"merged_script": "print('First run')"}
        dummy_execution_info = {"info": "first run"}
        dummy_graph = DummyGraph(dummy_final_state, dummy_execution_info)
        instance.graph = dummy_graph
        result1 = instance.run()
        # Modify dummy graph's state for the second run.
        dummy_graph.final_state["merged_script"] = "print('Second run')"
        dummy_graph.execution_info = {"info": "second run"}
        result2 = instance.run()
        assert result1 == "print('First run')"
        assert result2 == "print('Second run')"



"""
This module contains the functions that generate prompts for various types of code error analysis.

Functions:
- syntax_focused_analysis: Focuses on syntax-related errors in the generated code.
- execution_focused_analysis: Focuses on execution-related errors,
including generated code and HTML analysis.
- validation_focused_analysis: Focuses on validation-related errors,
considering JSON schema and execution result.
- semantic_focused_analysis: Focuses on semantic differences in
generated code based on a comparison result.
"""

import json
from typing import Any, Dict

from langchain.prompts import PromptTemplate
from langchain_core.output_parsers import StrOutputParser

from ..prompts import (
    TEMPLATE_EXECUTION_ANALYSIS,
    TEMPLATE_SEMANTIC_ANALYSIS,
    TEMPLATE_SYNTAX_ANALYSIS,
    TEMPLATE_VALIDATION_ANALYSIS,
)


def syntax_focused_analysis(state: dict, llm_model) -> str:
    """
    Analyzes the syntax errors in the generated code.

    Args:
        state (dict): Contains the 'generated_code' and 'errors' related to syntax.
        llm_model: The language model used for generating the analysis.

    Returns:
        str: The result of the syntax error analysis.
    """
    prompt = PromptTemplate(
        template=TEMPLATE_SYNTAX_ANALYSIS, input_variables=["generated_code", "errors"]
    )
    chain = prompt | llm_model | StrOutputParser()
    return chain.invoke(
        {"generated_code": state["generated_code"], "errors": state["errors"]["syntax"]}
    )


def execution_focused_analysis(state: dict, llm_model) -> str:
    """
    Analyzes the execution errors in the generated code and HTML code.

    Args:
        state (dict): Contains the 'generated_code', 'errors', 'html_code', and 'html_analysis'.
        llm_model: The language model used for generating the analysis.

    Returns:
        str: The result of the execution error analysis.
    """
    prompt = PromptTemplate(
        template=TEMPLATE_EXECUTION_ANALYSIS,
        input_variables=["generated_code", "errors", "html_code", "html_analysis"],
    )
    chain = prompt | llm_model | StrOutputParser()
    return chain.invoke(
        {
            "generated_code": state["generated_code"],
            "errors": state["errors"]["execution"],
            "html_code": state["html_code"],
            "html_analysis": state["html_analysis"],
        }
    )


def validation_focused_analysis(state: dict, llm_model) -> str:
    """
    Analyzes the validation errors in the generated code based on a JSON schema.

    Args:
        state (dict): Contains the 'generated_code', 'errors',
        'json_schema', and 'execution_result'.
        llm_model: The language model used for generating the analysis.

    Returns:
        str: The result of the validation error analysis.
    """
    prompt = PromptTemplate(
        template=TEMPLATE_VALIDATION_ANALYSIS,
        input_variables=["generated_code", "errors", "json_schema", "execution_result"],
    )
    chain = prompt | llm_model | StrOutputParser()
    return chain.invoke(
        {
            "generated_code": state["generated_code"],
            "errors": state["errors"]["validation"],
            "json_schema": state["json_schema"],
            "execution_result": state["execution_result"],
        }
    )


def semantic_focused_analysis(
    state: dict, comparison_result: Dict[str, Any], llm_model
) -> str:
    """
    Analyzes the semantic differences in the generated code based on a comparison result.

    Args:
        state (dict): Contains the 'generated_code'.
        comparison_result (Dict[str, Any]): Contains
        'differences' and 'explanation' of the comparison.
        llm_model: The language model used for generating the analysis.

    Returns:
        str: The result of the semantic error analysis.
    """
    prompt = PromptTemplate(
        template=TEMPLATE_SEMANTIC_ANALYSIS,
        input_variables=["generated_code", "differences", "explanation"],
    )
    chain = prompt | llm_model | StrOutputParser()
    return chain.invoke(
        {
            "generated_code": state["generated_code"],
            "differences": json.dumps(comparison_result["differences"], indent=2),
            "explanation": comparison_result["explanation"],
        }
    )



"""
This module contains the telemetry module for the scrapegraphai package.
"""

from .telemetry import disable_telemetry, log_event, log_graph_execution

__all__ = [
    "disable_telemetry",
    "log_event",
    "log_graph_execution",
]



"""
Module for counting tokens and splitting text into chunks
"""

from .tokenizers.tokenizer_openai import num_tokens_openai


def num_tokens_calculus(string: str) -> int:
    """
    Returns the number of tokens in a text string.
    """

    num_tokens_fn = num_tokens_openai

    num_tokens = num_tokens_fn(string)
    return num_tokens



from unittest.mock import Mock, patch

import pytest
from pydantic import BaseModel, Field

from scrapegraphai.graphs.json_scraper_graph import JSONScraperGraph


class TestJSONScraperGraph:
    @pytest.fixture
    def mock_llm_model(self):
        return Mock()

    @pytest.fixture
    def mock_embedder_model(self):
        return Mock()

    @patch("scrapegraphai.graphs.json_scraper_graph.FetchNode")
    @patch("scrapegraphai.graphs.json_scraper_graph.GenerateAnswerNode")
    @patch.object(JSONScraperGraph, "_create_llm")
    def test_json_scraper_graph_with_directory(
        self,
        mock_create_llm,
        mock_generate_answer_node,
        mock_fetch_node,
        mock_llm_model,
        mock_embedder_model,
    ):
        """
        Test JSONScraperGraph with a directory of JSON files.
        This test checks if the graph correctly handles multiple JSON files input
        and processes them to generate an answer.
        """
        # Mock the _create_llm method to return a mock LLM model
        mock_create_llm.return_value = mock_llm_model

        # Mock the execute method of BaseGraph
        with patch(
            "scrapegraphai.graphs.json_scraper_graph.BaseGraph.execute"
        ) as mock_execute:
            mock_execute.return_value = (
                {"answer": "Mocked answer for multiple JSON files"},
                {},
            )

            # Create a JSONScraperGraph instance
            graph = JSONScraperGraph(
                prompt="Summarize the data from all JSON files",
                source="path/to/json/directory",
                config={"llm": {"model": "test-model", "temperature": 0}},
                schema=BaseModel,
            )

            # Set mocked embedder model
            graph.embedder_model = mock_embedder_model

            # Run the graph
            result = graph.run()

            # Assertions
            assert result == "Mocked answer for multiple JSON files"
            assert graph.input_key == "json_dir"
            mock_execute.assert_called_once_with(
                {
                    "user_prompt": "Summarize the data from all JSON files",
                    "json_dir": "path/to/json/directory",
                }
            )
            mock_fetch_node.assert_called_once()
            mock_generate_answer_node.assert_called_once()
            mock_create_llm.assert_called_once_with(
                {"model": "test-model", "temperature": 0}
            )

    @pytest.fixture
    def mock_llm_model(self):
        return Mock()

    @pytest.fixture
    def mock_embedder_model(self):
        return Mock()

    @patch("scrapegraphai.graphs.json_scraper_graph.FetchNode")
    @patch("scrapegraphai.graphs.json_scraper_graph.GenerateAnswerNode")
    @patch.object(JSONScraperGraph, "_create_llm")
    def test_json_scraper_graph_with_single_file(
        self,
        mock_create_llm,
        mock_generate_answer_node,
        mock_fetch_node,
        mock_llm_model,
        mock_embedder_model,
    ):
        """
        Test JSONScraperGraph with a single JSON file.
        This test checks if the graph correctly handles a single JSON file input
        and processes it to generate an answer.
        """
        # Mock the _create_llm method to return a mock LLM model
        mock_create_llm.return_value = mock_llm_model

        # Mock the execute method of BaseGraph
        with patch(
            "scrapegraphai.graphs.json_scraper_graph.BaseGraph.execute"
        ) as mock_execute:
            mock_execute.return_value = (
                {"answer": "Mocked answer for single JSON file"},
                {},
            )

            # Create a JSONScraperGraph instance with a single JSON file
            graph = JSONScraperGraph(
                prompt="Analyze the data from the JSON file",
                source="path/to/single/file.json",
                config={"llm": {"model": "test-model", "temperature": 0}},
                schema=BaseModel,
            )

            # Set mocked embedder model
            graph.embedder_model = mock_embedder_model

            # Run the graph
            result = graph.run()

            # Assertions
            assert result == "Mocked answer for single JSON file"
            assert graph.input_key == "json"
            mock_execute.assert_called_once_with(
                {
                    "user_prompt": "Analyze the data from the JSON file",
                    "json": "path/to/single/file.json",
                }
            )
            mock_fetch_node.assert_called_once()
            mock_generate_answer_node.assert_called_once()
            mock_create_llm.assert_called_once_with(
                {"model": "test-model", "temperature": 0}
            )

    @patch("scrapegraphai.graphs.json_scraper_graph.FetchNode")
    @patch("scrapegraphai.graphs.json_scraper_graph.GenerateAnswerNode")
    @patch.object(JSONScraperGraph, "_create_llm")
    def test_json_scraper_graph_no_answer_found(
        self,
        mock_create_llm,
        mock_generate_answer_node,
        mock_fetch_node,
        mock_llm_model,
        mock_embedder_model,
    ):
        """
        Test JSONScraperGraph when no answer is found.
        This test checks if the graph correctly handles the scenario where no answer is generated,
        ensuring it returns the default "No answer found." message.
        """
        # Mock the _create_llm method to return a mock LLM model
        mock_create_llm.return_value = mock_llm_model

        # Mock the execute method of BaseGraph to return an empty answer
        with patch(
            "scrapegraphai.graphs.json_scraper_graph.BaseGraph.execute"
        ) as mock_execute:
            mock_execute.return_value = ({}, {})  # Empty state and execution info

            # Create a JSONScraperGraph instance
            graph = JSONScraperGraph(
                prompt="Query that produces no answer",
                source="path/to/empty/file.json",
                config={"llm": {"model": "test-model", "temperature": 0}},
                schema=BaseModel,
            )

            # Set mocked embedder model
            graph.embedder_model = mock_embedder_model

            # Run the graph
            result = graph.run()

            # Assertions
            assert result == "No answer found."
            assert graph.input_key == "json"
            mock_execute.assert_called_once_with(
                {
                    "user_prompt": "Query that produces no answer",
                    "json": "path/to/empty/file.json",
                }
            )
            mock_fetch_node.assert_called_once()
            mock_generate_answer_node.assert_called_once()
            mock_create_llm.assert_called_once_with(
                {"model": "test-model", "temperature": 0}
            )

    @pytest.fixture
    def mock_llm_model(self):
        return Mock()

    @pytest.fixture
    def mock_embedder_model(self):
        return Mock()

    @patch("scrapegraphai.graphs.json_scraper_graph.FetchNode")
    @patch("scrapegraphai.graphs.json_scraper_graph.GenerateAnswerNode")
    @patch.object(JSONScraperGraph, "_create_llm")
    def test_json_scraper_graph_with_custom_schema(
        self,
        mock_create_llm,
        mock_generate_answer_node,
        mock_fetch_node,
        mock_llm_model,
        mock_embedder_model,
    ):
        """
        Test JSONScraperGraph with a custom schema.
        This test checks if the graph correctly handles a custom schema input
        and passes it to the GenerateAnswerNode.
        """

        # Define a custom schema
        class CustomSchema(BaseModel):
            name: str = Field(..., description="Name of the attraction")
            description: str = Field(..., description="Description of the attraction")

        # Mock the _create_llm method to return a mock LLM model
        mock_create_llm.return_value = mock_llm_model

        # Mock the execute method of BaseGraph
        with patch(
            "scrapegraphai.graphs.json_scraper_graph.BaseGraph.execute"
        ) as mock_execute:
            mock_execute.return_value = (
                {"answer": "Mocked answer with custom schema"},
                {},
            )

            # Create a JSONScraperGraph instance with a custom schema
            graph = JSONScraperGraph(
                prompt="List attractions in Chioggia",
                source="path/to/chioggia.json",
                config={"llm": {"model": "test-model", "temperature": 0}},
                schema=CustomSchema,
            )

            # Set mocked embedder model
            graph.embedder_model = mock_embedder_model

            # Run the graph
            result = graph.run()

            # Assertions
            assert result == "Mocked answer with custom schema"
            assert graph.input_key == "json"
            mock_execute.assert_called_once_with(
                {
                    "user_prompt": "List attractions in Chioggia",
                    "json": "path/to/chioggia.json",
                }
            )
            mock_fetch_node.assert_called_once()
            mock_generate_answer_node.assert_called_once()

            # Check if the custom schema was passed to GenerateAnswerNode
            generate_answer_node_call = mock_generate_answer_node.call_args[1]
            assert generate_answer_node_call["node_config"]["schema"] == CustomSchema

            mock_create_llm.assert_called_once_with(
                {"model": "test-model", "temperature": 0}
            )



"""
This module provides helper functions and utilities for the ScrapeGraphAI application.
"""

from .models_tokens import models_tokens
from .nodes_metadata import nodes_metadata
from .robots import robots_dictionary
from .schemas import graph_schema

__all__ = [
    "models_tokens",
    "nodes_metadata",
    "robots_dictionary",
    "graph_schema",
]



"""
This module implements the Omni Scraper Graph for the ScrapeGraphAI application.
"""

from typing import Optional, Type

from pydantic import BaseModel

from ..models import OpenAIImageToText
from ..nodes import FetchNode, GenerateAnswerOmniNode, ImageToTextNode, ParseNode
from .abstract_graph import AbstractGraph
from .base_graph import BaseGraph


class OmniScraperGraph(AbstractGraph):
    """
    OmniScraper is a scraping pipeline that automates the process of
    extracting information from web pages
    using a natural language model to interpret and answer prompts.

    Attributes:
        prompt (str): The prompt for the graph.
        source (str): The source of the graph.
        config (dict): Configuration parameters for the graph.
        schema (BaseModel): The schema for the graph output.
        llm_model: An instance of a language model client, configured for generating answers.
        embedder_model: An instance of an embedding model client,
        configured for generating embeddings.
        verbose (bool): A flag indicating whether to show print statements during execution.
        headless (bool): A flag indicating whether to run the graph in headless mode.
        max_images (int): The maximum number of images to process.

    Args:
        prompt (str): The prompt for the graph.
        source (str): The source of the graph.
        config (dict): Configuration parameters for the graph.
        schema (BaseModel): The schema for the graph output.

    Example:
        >>> omni_scraper = OmniScraperGraph(
        ...     "List me all the attractions in Chioggia and describe their pictures.",
        ...     "https://en.wikipedia.org/wiki/Chioggia",
        ...     {"llm": {"model": "openai/gpt-4o"}}
        ... )
        >>> result = omni_scraper.run()
        )
    """

    def __init__(
        self,
        prompt: str,
        source: str,
        config: dict,
        schema: Optional[Type[BaseModel]] = None,
    ):
        self.max_images = 5 if config is None else config.get("max_images", 5)

        super().__init__(prompt, config, source, schema)

        self.input_key = "url" if source.startswith("http") else "local_dir"

    def _create_graph(self) -> BaseGraph:
        """
        Creates the graph of nodes representing the workflow for web scraping.

        Returns:
            BaseGraph: A graph instance representing the web scraping workflow.
        """

        fetch_node = FetchNode(
            input="url | local_dir",
            output=["doc"],
            node_config={
                "loader_kwargs": self.config.get("loader_kwargs", {}),
                "storage_state": self.config.get("storage_state"),
            },
        )

        parse_node = ParseNode(
            input="doc & (url | local_dir)",
            output=["parsed_doc", "link_urls", "img_urls"],
            node_config={
                "chunk_size": self.model_token,
                "parse_urls": True,
                "llm_model": self.llm_model,
            },
        )

        image_to_text_node = ImageToTextNode(
            input="img_urls",
            output=["img_desc"],
            node_config={
                "llm_model": OpenAIImageToText(self.config["llm"]),
                "max_images": self.max_images,
            },
        )

        generate_answer_omni_node = GenerateAnswerOmniNode(
            input="user_prompt & (relevant_chunks | parsed_doc | doc) & img_desc",
            output=["answer"],
            node_config={
                "llm_model": self.llm_model,
                "additional_info": self.config.get("additional_info"),
                "schema": self.schema,
            },
        )

        return BaseGraph(
            nodes=[
                fetch_node,
                parse_node,
                image_to_text_node,
                generate_answer_omni_node,
            ],
            edges=[
                (fetch_node, parse_node),
                (parse_node, image_to_text_node),
                (image_to_text_node, generate_answer_omni_node),
            ],
            entry_point=fetch_node,
            graph_name=self.__class__.__name__,
        )

    def run(self) -> str:
        """
        Executes the scraping process and returns the answer to the prompt.

        Returns:
            str: The answer to the prompt.
        """

        inputs = {"user_prompt": self.prompt, self.input_key: self.source}
        self.final_state, self.execution_info = self.graph.execute(inputs)

        return self.final_state.get("answer", "No answer found.")



"""
Module for showing how PDFScraper multi works
"""

import json
import os

from dotenv import load_dotenv

from scrapegraphai.graphs import JSONScraperMultiGraph

load_dotenv()

openai_key = os.getenv("OPENAI_APIKEY")

graph_config = {
    "llm": {
        "api_key": openai_key,
        "model": "openai/gpt-4o",
    }
}

FILE_NAME = "inputs/example.json"
curr_dir = os.path.dirname(os.path.realpath(__file__))
file_path = os.path.join(curr_dir, FILE_NAME)

with open(file_path, "r", encoding="utf-8") as file:
    text = file.read()

sources = [text, text]

multiple_search_graph = JSONScraperMultiGraph(
    prompt="List me all the authors, title and genres of the books",
    source=sources,
    schema=None,
    config=graph_config,
)

result = multiple_search_graph.run()
print(json.dumps(result, indent=4))



"""
browserbase integration module
"""

import asyncio
from typing import List


def browser_base_fetch(
    api_key: str,
    project_id: str,
    link: List[str],
    text_content: bool = True,
    async_mode: bool = False,
) -> List[str]:
    """
    BrowserBase Fetch

    This module provides an interface to the BrowserBase API.

    Args:
        api_key (str): The API key provided by BrowserBase.
        project_id (str): The ID of the project on BrowserBase where you want to fetch data from.
        link (List[str]): The URLs or links that you want to fetch data from.
        text_content (bool): Whether to return only the text content (True) or the full HTML (False).
        async_mode (bool): Whether to run the function asynchronously (True) or synchronously (False).

    Returns:
        List[str]: The results of the loading operations.
    """
    try:
        from browserbase import Browserbase
    except ImportError:
        raise ImportError(
            "The browserbase module is not installed. Please install it using `pip install browserbase`."
        )

    # Initialize client with API key
    browserbase = Browserbase(api_key=api_key)

    # Create session with project ID
    session = browserbase.sessions.create(project_id=project_id)

    result = []

    async def _async_fetch_link(url):
        return await asyncio.to_thread(session.load, url, text_content=text_content)

    if async_mode:

        async def _async_browser_base_fetch():
            for url in link:
                result.append(await _async_fetch_link(url))
            return result

        result = asyncio.run(_async_browser_base_fetch())
    else:
        for url in link:
            result.append(session.load(url, text_content=text_content))

    return result



"""
Module for minimizing the code
"""

import json
import re
from urllib.parse import urljoin

from bs4 import BeautifulSoup, Comment
from minify_html import minify


def extract_from_script_tags(soup):
    script_content = []

    for script in soup.find_all("script"):
        content = script.string
        if content:
            try:
                json_pattern = r"(?:const|let|var)?\s*\w+\s*=\s*({[\s\S]*?});?$"
                json_matches = re.findall(json_pattern, content)

                for potential_json in json_matches:
                    try:
                        parsed = json.loads(potential_json)
                        if parsed:
                            script_content.append(
                                f"JSON data from script: {json.dumps(parsed, indent=2)}"
                            )
                    except json.JSONDecodeError:
                        pass

                if "window." in content or "document." in content:
                    data_pattern = r"(?:window|document)\.(\w+)\s*=\s*([^;]+);"
                    data_matches = re.findall(data_pattern, content)

                    for var_name, var_value in data_matches:
                        script_content.append(
                            f"Dynamic data - {var_name}: {var_value.strip()}"
                        )
            except Exception:
                if len(content) < 1000:
                    script_content.append(f"Script content: {content.strip()}")

    return "\n\n".join(script_content)


def cleanup_html(html_content: str, base_url: str) -> str:
    """
    Processes HTML content by removing unnecessary tags,
    minifying the HTML, and extracting the title and body content.

    Args:
        html_content (str): The HTML content to be processed.

    Returns:
        str: A string combining the parsed title and the minified body content.
        If no body content is found, it indicates so.

    Example:
        >>> html_content = "<html><head><title>Example</title></head><body><p>Hello World!</p></body></html>"
        >>> remover(html_content)
        'Title: Example, Body: <body><p>Hello World!</p></body>'

    This function is particularly useful for preparing HTML content for
    environments where bandwidth usage needs to be minimized.
    """

    soup = BeautifulSoup(html_content, "html.parser")

    title_tag = soup.find("title")
    title = title_tag.get_text() if title_tag else ""

    script_content = extract_from_script_tags(soup)

    for tag in soup.find_all("style"):
        tag.extract()

    link_urls = [
        urljoin(base_url, link["href"]) for link in soup.find_all("a", href=True)
    ]

    images = soup.find_all("img")
    image_urls = []
    for image in images:
        if "src" in image.attrs:
            if "http" not in image["src"]:
                image_urls.append(urljoin(base_url, image["src"]))
            else:
                image_urls.append(image["src"])

    body_content = soup.find("body")
    if body_content:
        minimized_body = minify(str(body_content))
        return title, minimized_body, link_urls, image_urls, script_content

    else:
        raise ValueError(
            f"""No HTML body content found, please try setting the 'headless'
                         flag to False in the graph configuration. HTML content: {html_content}"""
        )


def minify_html(html):
    """
    minify_html function
    """
    # Combine multiple regex operations into one for better performance
    patterns = [
        (r"<!--.*?-->", "", re.DOTALL),
        (r">\s+<", "><", 0),
        (r"\s+>", ">", 0),
        (r"<\s+", "<", 0),
        (r"\s+", " ", 0),
        (r"\s*=\s*", "=", 0),
    ]

    for pattern, repl, flags in patterns:
        html = re.sub(pattern, repl, html, flags=flags)

    return html.strip()


def reduce_html(html, reduction):
    """
    Reduces the size of the HTML content based on the specified level of reduction.

    Args:
        html (str): The HTML content to reduce.
        reduction (int): The level of reduction to apply to the HTML content.
            0: minification only,
            1: minification and removig unnecessary tags and attributes,
            2: minification, removig unnecessary tags and attributes,
            simplifying text content, removing of the head tag

    Returns:
        str: The reduced HTML content based on the specified reduction level.
    """
    if reduction == 0:
        return minify_html(html)

    soup = BeautifulSoup(html, "html.parser")

    for comment in soup.find_all(string=lambda text: isinstance(text, Comment)):
        comment.extract()

    for tag in soup(["style"]):
        tag.string = ""

    attrs_to_keep = ["class", "id", "href", "src", "type"]
    for tag in soup.find_all(True):
        for attr in list(tag.attrs):
            if attr not in attrs_to_keep:
                del tag[attr]

    if reduction == 1:
        return minify_html(str(soup))

    for tag in soup(["style"]):
        tag.decompose()

    body = soup.body
    if not body:
        return "No <body> tag found in the HTML"

    for tag in body.find_all(string=True):
        if tag.parent.name not in ["script"]:
            tag.replace_with(re.sub(r"\s+", " ", tag.strip())[:20])

    reduced_html = str(body)

    reduced_html = minify_html(reduced_html)

    return reduced_html



"""
This utility function trasfrom the pydantic schema into a more comprehensible schema.
"""


def transform_schema(pydantic_schema):
    """
    Transform the pydantic schema into a more comprehensible JSON schema.

    Args:
        pydantic_schema (dict): The pydantic schema.

    Returns:
        dict: The transformed JSON schema.
    """

    def process_properties(properties):
        result = {}
        for key, value in properties.items():
            if "type" in value:
                if value["type"] == "array":
                    if "$ref" in value["items"]:
                        ref_key = value["items"]["$ref"].split("/")[-1]
                        result[key] = [
                            process_properties(
                                pydantic_schema["$defs"][ref_key]["properties"]
                            )
                        ]
                    else:
                        result[key] = [value["items"]["type"]]
                else:
                    result[key] = {
                        "type": value["type"],
                        "description": value.get("description", ""),
                    }
            elif "$ref" in value:
                ref_key = value["$ref"].split("/")[-1]
                result[key] = process_properties(
                    pydantic_schema["$defs"][ref_key]["properties"]
                )
        return result

    return process_properties(pydantic_schema["properties"])
